import abc
import socket
import warnings

from typing import Callable, List, Optional, TypeVar

from pymaid.core import get_running_loop, Event

from .base import logger, TransportState


class Transport(abc.ABC):

    logger = logger
    ID = 0

    def _fatal_error(self, exc, message='Fatal error on transport'):
        # Should be called from exception handler only.
        if isinstance(exc, OSError):
            self.logger.debug("%r: %s", self, message, exc_info=True)
        else:
            self._loop.call_exception_handler({
                'message': message,
                'exception': exc,
                'socket': self._sock,
            })
        self._force_close(exc)


class SocketTransport(Transport):
    '''Base class for net transport.

    Wraps low level socket.
    Wrapped some attrs and methods, added some apis like `asyncio` `protocols`.
    '''

    WRAPPED_ATTRS = ('family', 'proto', 'timeout', 'type')
    WRAPPED_METHODS = ('getsockopt', 'setsockopt')
    STATE = TransportState

    BUFFER_FACTORY = bytearray

    def __init__(
        self,
        sock: socket.socket,
        *,
        on_open: Optional[List[Callable]] = None,
        on_close: Optional[List[Callable]] = None,
    ):
        self._loop = get_running_loop()
        self.wrap_sock(sock)
        self.__class__.ID += 1
        self.id = self.__class__.ID

        self.on_open = on_open or []
        self.on_close = on_close or []
        self.closed_event = Event()
        self.state = self.STATE.OPENED
        self.write_buffer = self.BUFFER_FACTORY()

    def wrap_sock(self, sock: socket.socket):
        self._sock = sock
        self._sock_fd = sock.fileno()
        self._wrap_sock(self.WRAPPED_ATTRS)
        self._wrap_sock(self.WRAPPED_METHODS)
        self.set_socket_default_options()

        self.peername = sock.getpeername()
        self.sockname = sock.getsockname()
        self._loop.add_reader(self._sock_fd, self._read_ready)

    def set_socket_default_options(self):
        pass

    def shutdown(self, reason=None):
        self._loop.remove_writer(self._sock_fd)
        self._sock.shutdown(socket.SHUT_WR)

    def close(self, exc=None):
        if self.state == self.STATE.CLOSING:
            return
        self.state = self.STATE.CLOSING
        loop = self._loop
        loop.remove_reader(self._sock_fd)

        if not self.write_buffer:
            loop.remove_writer(self._sock_fd)
            # loop.call_soon(self._finnal_close, None)
            self._finnal_close(exc)

    async def wait_closed(self):
        await self.closed_event.wait()

    def _wrap_sock(self, keys: List[str]):
        for key in keys:
            setattr(self, key, getattr(self._sock, key))

    def _read_ready(self):
        raise NotImplementedError('_read_ready')

    def _force_close(self, exc):
        if self.state == self.STATE.CLOSED:
            return
        if self.write_buffer:
            self.write_buffer.clear()
            self._loop.remove_writer(self._sock_fd)
        self._loop.remove_reader(self._sock_fd)
        # self._loop.call_soon(self._finnal_close, exc)
        self._finnal_close(exc)

    def _finnal_close(self, exc=None):
        self.logger.info(f'{self!r} final close {exc=}')
        self.state = self.STATE.CLOSED
        for cb in self.on_close:
            cb(self, exc)
        self._sock.close()
        self._sock = None
        self._loop = None
        del self.on_open
        del self.on_close
        self.closed_event.set()

    def __del__(self, _warn=warnings.warn):
        if getattr(self, '_sock', None):
            _warn(
                f'unclosed transport {self.__class__.__name__}',
                ResourceWarning,
                source=self,
            )
            self._sock.close()
            self._sock = None

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} id={self.id} sock_fd={self._sock_fd} '
            f'state={self.state.name} initiative={self.initiative} '
            f'sockname={self.sockname} peername={self.peername}'
            f'>'
        )


TransportType = TypeVar('TransportType', bound=Transport)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymaid/ext/monitor/monitor.proto
"""Generated protocol buffer code."""
from pymaid.rpc.pb import pymaid_pb2 as pymaid_dot_rpc_dot_pb_dot_pymaid__pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name='pymaid/ext/monitor/monitor.proto',
    package='pymaid.ext.monitor',
    syntax='proto3',
    serialized_options=b'\220\001\001',
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n pymaid/ext/monitor/monitor.proto\x12\x12pymaid.ext.monitor\x1a\x1apymaid/rpc/pb/pymaid.proto\"\x06\n\x04Pong2G\n\x0eMonitorService\x12\x35\n\x04Ping\x12\x13.pymaid.rpc.pb.Void\x1a\x18.pymaid.ext.monitor.PongB\x03\x90\x01\x01\x62\x06proto3',
    dependencies=[pymaid_dot_rpc_dot_pb_dot_pymaid__pb2.DESCRIPTOR, ])


_PONG = _descriptor.Descriptor(
    name='Pong',
    full_name='pymaid.ext.monitor.Pong',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=84,
    serialized_end=90,
)

DESCRIPTOR.message_types_by_name['Pong'] = _PONG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Pong = _reflection.GeneratedProtocolMessageType('Pong', (_message.Message,), {
    'DESCRIPTOR': _PONG,
    '__module__': 'pymaid.ext.monitor.monitor_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.ext.monitor.Pong)
})
_sym_db.RegisterMessage(Pong)


DESCRIPTOR._options = None

_MONITORSERVICE = _descriptor.ServiceDescriptor(
    name='MonitorService',
    full_name='pymaid.ext.monitor.MonitorService',
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=92,
    serialized_end=163,
    methods=[
        _descriptor.MethodDescriptor(
            name='Ping',
            full_name='pymaid.ext.monitor.MonitorService.Ping',
            index=0,
            containing_service=None,
            input_type=pymaid_dot_rpc_dot_pb_dot_pymaid__pb2._VOID,
            output_type=_PONG,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ])
_sym_db.RegisterServiceDescriptor(_MONITORSERVICE)

DESCRIPTOR.services_by_name['MonitorService'] = _MONITORSERVICE

MonitorService = service_reflection.GeneratedServiceType('MonitorService', (_service.Service,), dict(
    DESCRIPTOR=_MONITORSERVICE,
    __module__='pymaid.ext.monitor.monitor_pb2'
))

MonitorService_Stub = service_reflection.GeneratedServiceStubType('MonitorService_Stub', (MonitorService,), dict(
    DESCRIPTOR=_MONITORSERVICE,
    __module__='pymaid.ext.monitor.monitor_pb2'
))


# @@protoc_insertion_point(module_scope)



import time

import pytest

from pymaid.core import run_in_threadpool, run_in_processpool


def sum_int(a: int, b: int) -> int:
    return a + b


def sleep(seconds):
    time.sleep(seconds)
    return time.time_ns()


@pytest.mark.asyncio
async def test_run_in_threadpool():
    assert await run_in_threadpool(sum_int, args=(1, 2)) == 3
    assert await run_in_threadpool(sum_int, kwargs={'a': 1, 'b': 2}) == 3

    now = time.time_ns()
    f1 = run_in_threadpool(sleep, args=(0.1,))
    f2 = run_in_threadpool(sleep, args=(0.1,))
    ts = time.time_ns()

    t1 = await f1
    t2 = await f2
    assert t1 > now + 0.1
    assert t2 > now + 0.1
    assert ts > now
    assert ts < t1
    assert ts < t2


@pytest.mark.asyncio
async def test_run_in_processpool():
    assert await run_in_processpool(sum_int, args=(1, 2)) == 3
    assert await run_in_processpool(sum_int, kwargs={'a': 1, 'b': 2}) == 3

    now = time.time_ns()
    f1 = run_in_processpool(sleep, args=(0.1,))
    f2 = run_in_processpool(sleep, args=(0.1,))
    ts = time.time_ns()

    t1 = await f1
    t2 = await f2
    assert t1 > now + 0.1
    assert t2 > now + 0.1
    assert ts > now
    assert ts < t1
    assert ts < t2



from typing import Callable, Coroutine, Optional, TypeVar

from pymaid.core import Event, Semaphore, Task
from pymaid.core import current_task, iscoroutine


class AioPool:

    def __init__(self, size: int = 1024, *, task_class: TypeVar(Task) = Task):
        if size <= 0:
            raise ValueError(f'size must be positive: {size}')

        if not issubclass(task_class, Task):
            raise TypeError(
                f'task_class expected to be subclass of Task, '
                f'got: {task_class}'
            )
        self.task_class = task_class

        self.size = size
        self.semaphore = Semaphore(size)
        self.tasks = set()
        self.empty_event = Event()

        self.has_shutdown = False
        self.executed_count = 0

    @property
    def is_empty(self):
        return len(self.tasks) == 0

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_tpye, exc_value, exc_tb):
        await self.join()

    async def _run(self, coro, callback=None):
        try:
            result = await coro
            if callback is not None:
                callback(result)
            return result
        finally:
            self.executed_count += 1
            self.task_done(current_task())

    async def _spawn(self, coro, callback=None):
        await self.semaphore.acquire()
        try:
            return await self._run(coro, callback)
        finally:
            self.semaphore.release()

    def task_done(self, task):
        assert task in self.tasks, 'task not belong to this pool'
        self.tasks.remove(task)
        if self.is_empty:
            self.notify_empty()

    async def spawn(
        self, coro: Coroutine, callback: Optional[Callable] = None
    ) -> Task:
        '''Submit coroutine to the pool, waiting for pool space.

        coroutine is executed in pool when pool space is available.
        '''
        if self.has_shutdown:
            raise RuntimeError('cannot call spawn after shutdown')

        if not iscoroutine(coro):
            raise TypeError(f'coro expected to be coroutine, got: {coro}')

        if callback is not None and not callable(callback):
            raise TypeError(
                f'callback expected to be callable, got: {callback}'
            )

        await self.semaphore.acquire()
        task = self.task_class(self._run(coro, callback=callback))
        task.add_done_callback(lambda t: self.semaphore.release())
        self.tasks.add(task)
        return task

    def submit(
        self, coro: Coroutine, callback: Optional[Callable] = None
    ) -> Task:
        '''Submit coroutine to the pool, without waiting for pool space.

        coroutine is executed in pool when pool space is available.
        '''
        if self.has_shutdown:
            raise RuntimeError('cannot call submit after shutdown')

        if not iscoroutine(coro):
            raise TypeError(f'coro expected to be coroutine, got: {coro}')

        if callback is not None and not callable(callback):
            raise TypeError(
                f'callback expected to be callable, got: {callback}'
            )

        task = self.task_class(self._spawn(coro, callback=callback))
        self.tasks.add(task)
        return task

    async def shutdown(self, wait=True):
        self.has_shutdown = True
        if wait:
            await self.join()

    async def join(self):
        if current_task() in self.tasks:
            raise RuntimeError(
                'cannot call join within task spawned by the pool, '
                'it will cause deadlock'
            )

        if not self.is_empty:
            await self.empty_event.wait()

    def notify_empty(self):
        '''Wake up join waiters, reset empty_event for reusability.'''
        self.empty_event.set()
        self.empty_event.clear()



from bisect import bisect
from hashlib import md5


primes = (
    2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71,
    73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151,
    157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233,
    239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317,
    331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419,
    421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503,
    509, 521, 523, 541, 547, 557, 563, 569, 571, 577, 587, 593, 599, 601, 607,
    613, 617, 619, 631, 641, 643, 647, 653, 659, 661, 673, 677, 683, 691, 701,
    709, 719, 727, 733, 739, 743, 751, 757, 761, 769, 773, 787, 797, 809, 811,
    821, 823, 827, 829, 839, 853, 857, 859, 863, 877, 881, 883, 887, 907, 911,
    919, 929, 937, 941, 947, 953, 967, 971, 977, 983, 991, 997
)


def md5_hash_func(key):
    return int(md5(key.encode('utf-8')).hexdigest(), 16)


class HashNode(object):

    def __init__(self, key, weight=16, enabled=True):
        self.key = key
        self.hashed_key = md5_hash_func(key)
        self.weight = weight
        self.enabled = enabled

    def __eq__(self, other):
        if not isinstance(other, HashNode):
            return NotImplemented
        return self.hashed_key == other.hashed_key

    def __ne__(self, other):
        return self != other

    def __hash__(self):
        return self.hashed_key


class BaseHashManager(object):

    def __init__(self, name, hash_func=md5_hash_func):
        self.name = name
        self.objects = {}
        self.nodes = []
        self.hash_func = hash_func

    def add_node(self, node):
        if node.key in self.objects:
            return
        self.objects[node.key] = node
        node._hash_manager = self
        if node.enabled:
            self.nodes.append(node)
            self.rehash()

    def add_nodes(self, nodes):
        for node in nodes:
            if node.key not in self.objects:
                self.objects[node.key] = node
                node._hash_manager = self
                if node.enabled:
                    self.nodes.append(node)
        self.rehash()

    def remove_node(self, node):
        if node.key in self.objects:
            del self.objects[node.key]
            node._hash_manager = None
            del node._hash_manager
        if node in self.nodes:
            self.nodes.remove(node)
        self.rehash()

    def enable_node(self, key):
        if key not in self.objects:
            return
        node = self.objects[key]
        node.enabled = True
        if node not in self.nodes:
            self.nodes.append(node)
            self.rehash()

    def disable_node(self, key):
        if key not in self.objects:
            return
        node = self.objects[key]
        node.enabled = False
        if node in self.nodes:
            self.nodes.remove(node)
            self.rehash()

    def reset(self):
        self.objects.clear()
        del self.nodes[:]

    def filter(self, keys):
        # return a copy of this manager with filtered keys
        obj = self.__class__(self.name, self.hash_func)
        obj.objects = {
            o.key: o for o in self.objects.values() if o.key in keys
        }
        obj.nodes = [node for node in self.nodes if node.key in keys]
        obj.rehash()
        return obj

    def rehash(self):
        raise NotImplementedError

    def get_node(self, key):
        raise NotImplementedError

    def clone(self):
        # return a copy of this manager
        raise NotImplementedError

    def __str__(self):
        return f'<{self.__class__.__name__}: {self.name}>'


class HashRing(BaseHashManager):

    def __init__(self, name, hash_func=md5_hash_func):
        super(HashRing, self).__init__(name, hash_func)
        self.lookup_table = {}
        self.sorted_keys = []

    def rehash(self):
        self.lookup_table = {}
        self.sorted_keys = []

        if not self.nodes:
            return

        hash_func = self.hash_func
        lookup_table = self.lookup_table
        for node in self.nodes:
            key = node.key
            for idx in range(node.weight):
                virtual_key = hash_func(f'{key}-{idx}')
                if virtual_key in lookup_table:
                    # TODO: what to do?
                    continue
                lookup_table[virtual_key] = node
        self.sorted_keys = sorted(lookup_table.keys())

    def get_node(self, key):
        if not self.nodes:
            return

        virtual_key = self.hash_func(key)
        skeys = self.sorted_keys
        pos = bisect(skeys, virtual_key)
        return self.lookup_table[skeys[pos if pos < len(skeys) else 0]]

    def reset(self):
        super(HashRing, self).reset()
        self.lookup_table.clear()
        del self.sorted_keys[:]

    def clone(self):
        obj = self.__class__(self.name, self.hash_func)
        obj.objects = self.objects.copy()
        obj.nodes = self.nodes[:]
        obj.lookup_table = self.lookup_table.copy()
        obj.sorted_keys = self.sorted_keys[:]
        return obj


class MaglevHash(BaseHashManager):

    def __init__(self, name, hash_func=md5_hash_func, virtual_entry_count=16):
        super(MaglevHash, self).__init__(name, hash_func)
        self.virtual_entry_count = virtual_entry_count
        self.lookup_table = []

    def rehash(self):
        self.lookup_table = []

        if not self.nodes:
            return

        permutation = []
        hash_func = self.hash_func
        entry_count = len(self.nodes) * self.virtual_entry_count
        pos = bisect(primes, entry_count)
        entry_count = primes[pos if pos < len(primes) else -1]
        for node in self.nodes:
            key = node.key
            offset = hash_func('cat' + key) % entry_count
            skip = (hash_func('lee' + key) % (entry_count - 1)) + 1
            permutation.append([
                (offset + idx * skip) % entry_count
                for idx in range(entry_count)
            ])

        nexts = [0] * len(self.nodes)
        entries = self.lookup_table = [-1] * entry_count

        n = 0
        while 1:
            for idx in range(len(self.nodes)):
                c = permutation[idx][nexts[idx]]
                while entries[c] != -1:
                    nexts[idx] += 1
                    c = permutation[idx][nexts[idx]]
                entries[c] = idx
                nexts[idx] += 1
                n += 1

                if n == entry_count:
                    return

    def get_node(self, key):
        if not self.nodes:
            return
        key = self.hash_func('cat' + key)
        return self.nodes[self.lookup_table[key % len(self.lookup_table)]]

    def reset(self):
        super(MaglevHash, self).reset()
        del self.lookup_table[:]

    def clone(self):
        obj = self.__class__(self.name, self.hash_func)
        obj.objects = self.objects.copy()
        obj.nodes = self.nodes[:]
        obj.lookup_table = self.lookup_table[:]
        obj.virtual_entry_count = self.virtual_entry_count
        return obj



import os
import sys
import threading
from contextlib import contextmanager

from pymaid import LifoQueue, PriorityQueue, QueueFull, QueueEmpty


class ConnectionPool(object):
    '''Inspired by redis.ConnectionPool'''

    queue_class = LifoQueue
    empty_item = None

    def __init__(self, name, size=50, init_count=0, channel=None,
                 **connection_kwargs):
        '''Create a blocking connection pool.

        Use ``size`` to increase / decrease the pool size::

        By default, connections will be created by channel.acquire method.

        Any additional keyword arguments are passed to the channel.acquire.
        '''
        if not isinstance(size, int) or size < 0 or size > 100:
            raise ValueError('size must be 0 < size <= 100')

        self.name = name
        self.size = size
        self.init_count = init_count
        self.channel = channel
        self.connection_kwargs = connection_kwargs
        self.reset()

    def _checkpid(self):
        if self.pid != os.getpid():
            with self._check_lock:
                if self.pid == os.getpid():
                    # another thread already did the work while we waited
                    # on the lock.
                    return
                self.disconnect()
                self.reset()

    def reset(self):
        self.pid = os.getpid()
        self._check_lock = threading.Lock()

        # Create and fill up a thread safe queue with ``empty_item`` values.
        self.pool = self.queue_class(self.size)
        while 1:
            try:
                self.pool.put_nowait(self.empty_item)
            except QueueFull:
                break

        # Keep a list of actual connection instances so that we can
        # disconnect them later.
        self._connections = []
        if self.init_count:
            self.initpool(self.init_count)

    def item_getter(self, item):
        return item

    def item_putter(self, conn):
        return conn

    def initpool(self, init_count):
        if self._connections:
            # do not init pool if there are connections
            return
        if init_count > self.size:
            raise ValueError('init_count should not greater than pool size')
        if not self.channel:
            raise ValueError('calling initpool with channel is None')
        self.init_count = init_count
        try:
            for _ in range(init_count):
                self.pool.get_nowait()
        except QueueEmpty:
            pass
        try:
            for _ in range(init_count):
                self.pool.put_nowait(self.item_putter(self.make_connection()))
        except QueueFull:
            pass

    def get_connection(self, timeout=None):
        '''
        Get a connection from the pool, blocking for ``timeout`` util
        a connection is available from the pool.
        '''
        self._checkpid()
        # will raise QueueEmpty if timeout is not None
        item = self.pool.get(block=True, timeout=timeout)
        if item is self.empty_item:
            return self.make_connection()
        else:
            return self.item_getter(item)

    @contextmanager
    def get_autorelease_connection(self, timeout=None):
        conn = self.get_connection(timeout)
        try:
            yield conn
        finally:
            self.release(conn)

    def make_connection(self):
        '''Create a new connection'''
        connection = self.channel.acquire(**self.connection_kwargs)
        connection.pid = os.getpid()

        def close(conn, reason=None, reset=None):
            self._connections.remove(conn)
            item = self.item_putter(conn)
            if item in self.pool.queue:
                self.pool.queue.remove(item)
            del connection.pid
            try:
                self.pool.put_nowait(self.empty_item)
            except QueueFull:
                pass

        connection.add_close_cb(close)
        self._connections.append(connection)
        return connection

    def release(self, connection):
        '''Releases the connection back to the pool'''
        self._checkpid()
        if connection.pid != self.pid:
            return

        if connection.is_closed:
            if connection in self._connections:
                self._connections.remove(connection)
            connection = None
            item = self.empty_item
        else:
            item = self.item_putter(connection)
        try:
            self.pool.put_nowait(item)
        except QueueFull:
            # perhaps the pool has been reset() after a fork? regardless,
            # we don't want this connection
            # should we close this conn?
            if connection:
                connection.close('useless')

    def disconnect(self, reason=None):
        '''Disconnects all connections in the pool'''
        for connection in self._connections[:]:
            connection.close(reason)

    def __repr__(self):
        return '[ConnectionPool|%s][connect_kwargs|%s][max|%d][created|%d]' % (
            self.name, self.connection_kwargs, self.size,
            len(self._connections)
        )
    __str__ = __repr__


class PriorityConnectionPool(ConnectionPool):

    queue_class = PriorityQueue
    empty_item = (sys.maxsize, None)

    def item_getter(self, item):
        return item[1] if item else item

    def item_putter(self, conn):
        return (conn.transmission_id, conn) if conn else conn



import pymaid
import pymaid.rpc.pb

from pymaid.ext.middleware import MiddlewareManager
from pymaid.ext.monitor import HeartbeatMiddleware

from examples.template import get_server_parser, parse_args


async def main(args):
    mm = MiddlewareManager([HeartbeatMiddleware(args.interval, args.retry)])
    await pymaid.rpc.pb.serve_stream(
        args.address, backlog=1024, services=[], middleware_manager=mm
    )


if __name__ == "__main__":
    parser = get_server_parser()
    parser.add_argument(
        'interval', type=int, default=10, help='heartbeat timeout in seconds'
    )
    parser.add_argument(
        'retry', type=int, default=3, help='retry before heartbeat timeout'
    )
    args = parse_args(parser)
    pymaid.run(main(args))



from .middleware import HeartbeatMiddleware
from .monitor_pb2 import MonitorService, MonitorService_Stub
from .service import MonitorServiceImpl


__all__ = [
    'MonitorService', 'MonitorService_Stub', 'MonitorServiceImpl',
    'HeartbeatMiddleware',
]



from contextlib import AsyncExitStack, ExitStack
from functools import wraps
from sys import _getframe as getframe

from pymaid.core import iscoroutinefunction

from .logger import logger_wrapper

__all__ = ['ObjectManager', 'get_ipaddress']


@logger_wrapper
class ObjectManager(object):

    def __init__(self, name):
        self.name = name
        self.objects = {}

    def add(self, pk, obj):
        self.logger.info('[%s][add|%r][%s]', self.name, pk, obj)
        assert pk not in self.objects, pk
        self.objects[pk] = obj
        obj._manager = self

    def has(self, pk):
        return pk in self.objects

    def get(self, pk):
        assert pk in self.objects, pk
        return self.objects[pk]

    def remove(self, pk):
        self.logger.info('[%s][remove|%r]', self.name, pk)
        assert pk in self.objects, pk
        obj = self.objects.pop(pk)
        obj._manager = None
        return obj


def get_ipaddress(ifname):
    import socket
    import struct
    import fcntl
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(
        fcntl.ioctl(
            s.fileno(), 0x8915, struct.pack('256s', ifname[:15])
        )[20:24]
    )


def listify(obj):
    if obj is None:
        return []
    return obj if isinstance(obj, (tuple, list)) else [obj]


def with_defer(func):
    '''Used to enable defer

    For example:

    @with_defer
    def func1():
        defer(print, 'done func1')

    if use with multiple decorators, put it in the top

    @with_defer
    @decorator2
    @decorator3
    def func2():
        defer(print, 'done func2')

    it will also automatically detect async function

    @with_defer
    async def func1():
        defer(coroutinefunction)
        defer(print, 'done func1')

    if use with multiple decorators, put it in the top

    @with_defer
    @decorator2
    @decorator3
    async def func2():
        defer(coroutinefunction)
        defer(print, 'done func2')
    '''
    if iscoroutinefunction(func):
        @wraps(func)
        async def _(*args, **kwargs):
            is_async = True  # noqa, used by defer below
            async with AsyncExitStack() as __stack__:  # noqa, used by defer below
                return await func(*args, **kwargs)
        return _
    else:
        @wraps(func)
        def _(*args, **kwargs):
            is_async = False  # noqa, used by defer below
            with ExitStack() as __stack__:  # noqa, used by defer below
                return func(*args, **kwargs)
        return _


def defer(func, *args, **kwargs):
    '''Register a callback to be called when exit a scope.

    The callbacks are executed in LIFO order.

    If a callback raises an Exception, it will be reraised after all
    callbacks are executed.

    If multiple callbacks raise Exceptions, it will chain the exceptions
    tracebacks and you will see output like:
        ...
        *During handling of the above exception, another exception occurred:*
        ...
        *During handling of the above exception, another exception occurred:*
        ...
    '''
    f = getframe(2)
    if '__stack__' not in f.f_locals:
        raise RuntimeError(
            'calling scope is missing the with_defer decorator, '
            'nor with_defer decorator is not the innermost one'
        )
    st = f.f_locals['__stack__']
    if iscoroutinefunction(func):
        if not f.f_locals['is_async']:
            raise ValueError(
                'defer coroutinefunction required within coroutinefunction'
            )
        st.push_async_callback(func, *args, **kwargs)
    elif callable(func):
        st.callback(func, *args, **kwargs)
    else:
        raise TypeError(f'a callable is required, got {func}')


del logger_wrapper



import pymaid
import pytest


@pytest.fixture
def event_loop():
    # as upgraded to pytest-asyncio == 0.14.0
    # it reset default event loop after every test
    # so, this is a workaround to use the loop as pymaid
    if not pymaid.conf.settings.pymaid.get('NO_UVLOOP'):
        import uvloop
        uvloop.install()
    loop = pymaid.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()



import logging
import logging.config
import inspect
import sys
import traceback
import warnings

from collections.abc import Mapping

__all__ = ['logger_wrapper', 'get_logger']

pymaid_logger = logging.getLogger('pymaid')


def configure_logging(settings, ns):
    '''Setup logging from pymaid.LOGGING and logging.LOGGING settings.

    Do nothing if `ns` not in (pymaid, logging)
    '''

    if ns not in ('pymaid', 'logging'):
        return

    # Route warnings through python logging
    logging.captureWarnings(True)
    # Allow DeprecationWarnings through the warnings filters
    warnings.simplefilter('default', DeprecationWarning)

    config = settings.get('LOGGING', {}, ns='pymaid')
    for key, value in settings.get('LOGGING', {}, ns='logging').items():
        if not isinstance(value, Mapping):
            config[key] = value
        else:
            config[key].update(value)
    if config:
        if ns == 'pymaid' and settings.pymaid.DEBUG:
            config['loggers']['pymaid']['level'] = 'DEBUG'
        logging.config.dictConfig(config)


def logger_wrapper(name=''):

    def _(cls):
        cls.logger = get_logger(name)
        return cls

    if isinstance(name, type):
        cls, name = name, name.__name__
        return _(cls)
    else:
        return _


def get_logger(name):
    return pymaid_logger.getChild(name)


def enable_debug_output(signum):

    def debug_output(sig, frame):
        sys.stderr.write('############## pymaid debug_output ##############\n')
        for tid, frame in sys._current_frames().iteritems():
            sys.stderr.write('\n')
            sys.stderr.write('thread: %s\n' % tid)
            sys.stderr.write(
                ''.join(traceback.format_list(traceback.extract_stack(frame)))
            )
            sys.stderr.write('%s' % str(inspect.getargvalues(frame)))

    from pymaid import backend
    backend.get_running_loop().add_signal_handler(signum, debug_output)



import enum

from pymaid.utils.logger import get_logger

logger = get_logger('net')


class TransportState(enum.IntEnum):

    UNKNOWN = 0
    OPENED = 10
    CLOSING = 20
    CLOSED = 30


class ChannelState(enum.IntEnum):

    CREATED = 0
    STARTED = 10
    PAUSED = 20
    CLOSING = 30
    CLOSED = 40



import abc
from sys import _getframe as getframe

from orjson import loads


class BaseEx(Exception):

    code = None
    message = 'BaseEx'

    def __init__(self, *args, **kwargs):
        if args or kwargs:
            self.message = self.message.format(*args, **kwargs)
        self.data = kwargs.get('data', {})


class Error(BaseEx):

    def __str__(self):
        return '[ERROR][code|{}][message|{}][data|{}]'.format(
            self.code, self.message, self.data
        )
    __repr__ = __str__

    def __bytes__(self):
        return '[ERROR][code|{}][message|{}][data|{}]'.format(
            self.code, self.message, self.data
        ).encode('utf-8')


class Warning(BaseEx):

    def __str__(self):
        return '[WARN][code|{}][message|{}][data|{}]'.format(
            self.code, self.message, self.data
        )
    __repr__ = __str__

    def __bytes__(self):
        return '[WARN][code|{}][message|{}][data|{}]'.format(
            self.code, self.message, self.data
        ).encode('utf-8')


class ErrorManager(metaclass=abc.ABCMeta):

    codes = {}
    managers = {}
    __fullname__ = ''

    @classmethod
    def add(cls, name, ex):
        if ex.code in cls.codes:
            raise ValueError(f'duplicated exception code: {ex.code}')
        cls.codes[ex.code] = ex
        setattr(cls, name, ex)

    @classmethod
    def add_manager(cls, name, manager):
        cls.managers[manager.__name__] = manager
        setattr(cls, name, manager)

    @classmethod
    def add_error(cls, name, message):
        frame = getframe(1)  # get caller frame
        if cls.__fullname__:
            fullname = f'{cls.__fullname__}.{name}'
        else:
            fullname = name
        error = type(
            name, (Error, cls),
            {
                'code': fullname,
                'message': message,
                '__module__': frame.f_locals.get('__name__', ''),
                '__fullname__': fullname,
            }
        )
        cls.add(name, error)
        cls.register(error)
        return error

    @classmethod
    def add_warning(cls, name, message):
        frame = getframe(1)  # get caller frame
        if cls.__fullname__:
            fullname = f'{cls.__fullname__}.{name}'
        else:
            fullname = name
        warning = type(
            name,
            (Warning, cls),
            {
                'code': fullname,
                'message': message,
                '__module__': frame.f_locals.get('__name__', ''),
                '__fullname__': fullname,
            }
        )
        cls.add(name, warning)
        cls.register(warning)
        return warning

    @classmethod
    def get_exception(cls, code):
        try:
            return cls.codes[code]
        except KeyError:
            for manager in cls.managers.values():
                ex = manager.get_exception(code)
                if ex is not None:
                    return ex

    @classmethod
    def assemble(cls, code, message, data):
        ex = ErrorManager.get_exception(code)
        if ex is None:
            ex = ErrorManager.add_warning('Unknown_%s' % code, message)
        ex = ex()
        ex.message = message
        if data:
            ex.data = loads(data)
        return ex

    @classmethod
    def create_manager(cls, name):
        frame = getframe(1)  # get caller frame
        kwargs = dict(ErrorManager.__dict__)
        kwargs['__module__'] = frame.f_locals.get('__name__', '')
        if cls.__fullname__:
            kwargs['__fullname__'] = f'{cls.__fullname__}.{name}'
        else:
            kwargs['__fullname__'] = name
        # added BaseEx as base class
        # for use specified :ErrorManager: in except clause
        manager = type(name, (cls, BaseEx), kwargs)
        manager.codes = {}
        manager.managers = {}
        cls.add_manager(name, manager)
        return manager



import socket

import pytest

from pymaid.core import getaddrinfo


@pytest.mark.asyncio
async def test_getaddrinfo_ipv4():
    infos = await getaddrinfo('localhost', 8888)
    assert infos
    # INET: stream, datagram, raw
    assert len(infos) == 3, infos

    info = infos[0]
    assert info[0] == socket.AF_INET
    assert info[1] == socket.SOCK_STREAM
    assert info[-1] == ('127.0.0.1', 8888)

    info = infos[1]
    assert info[0] == socket.AF_INET
    assert info[1] == socket.SOCK_DGRAM
    assert info[-1] == ('127.0.0.1', 8888)

    info = infos[2]
    assert info[0] == socket.AF_INET
    assert info[1] == socket.SOCK_RAW
    assert info[-1] == ('127.0.0.1', 8888)


@pytest.mark.asyncio
async def test_getaddrinfo_ipv6():
    infos = await getaddrinfo('::1', 8888)
    assert infos
    # INET: stream, datagram, raw
    assert len(infos) == 3, infos

    info = infos[0]
    assert info[0] == socket.AF_INET6
    assert info[1] == socket.SOCK_STREAM
    assert info[-1][:2] == ('::1', 8888)

    info = infos[1]
    assert info[0] == socket.AF_INET6
    assert info[1] == socket.SOCK_DGRAM
    assert info[-1][:2] == ('::1', 8888)

    info = infos[2]
    assert info[0] == socket.AF_INET6
    assert info[1] == socket.SOCK_RAW
    assert info[-1][:2] == ('::1', 8888)



import os
import socket

import pytest

from pymaid.core import sleep
from pymaid.net import dial_stream, serve_stream
from pymaid.net.raw import HAS_IPv6_FAMILY

from tests.common.models import _TestStreamChannel, _TestStream


@pytest.mark.asyncio
async def test_stream_channel_ipv4():
    server = await serve_stream(
        ('localhost', 8999),
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
        start_serving=True,
    )
    assert server

    client = await dial_stream(
        ('localhost', 8999),
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
    )
    stream = await client.acquire()
    assert stream.family == socket.AF_INET
    # localhost resolved to 127.0.0.1
    assert stream.peername == ('127.0.0.1', 8999)

    await stream.write(b'from pymaid')
    # skip one step to let data_received run
    await sleep(0)

    assert server.connected_stream
    assert server.connected_stream.received_data == b'from pymaid'

    stream.close()
    client.close()
    server.connected_stream.close()
    server.close()


@pytest.mark.skipif(not HAS_IPv6_FAMILY, reason='does not support ipv6')
@pytest.mark.asyncio
async def test_stream_channel_ipv6():
    server = await serve_stream(
        ('::1', 8999),
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
        start_serving=True,
    )
    assert server

    client = await dial_stream(
        ('::1', 8999),
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
    )
    stream = await client.acquire()
    assert stream.family == socket.AF_INET6
    assert stream.peername[:2] == ('::1', 8999)

    await stream.write(b'from pymaid')
    # skip one step to let data_received run
    await sleep(0)

    assert server.connected_stream
    assert server.connected_stream.received_data == b'from pymaid'

    stream.close()
    client.close()
    server.connected_stream.close()
    server.close()


@pytest.mark.skipif(
    not getattr(socket, 'AF_UNIX', None),
    reason='does not support unix domain sock'
)
@pytest.mark.asyncio
async def test_stream_channel_unix():
    server = await serve_stream(
        '/tmp/pymaid_test_ipv6.sock',
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
        start_serving=True,
    )
    assert server

    client = await dial_stream(
        '/tmp/pymaid_test_ipv6.sock',
        channel_class=_TestStreamChannel,
        transport_class=_TestStream,
    )
    stream = await client.acquire()
    assert stream.family == socket.AF_UNIX
    assert stream.peername == '/tmp/pymaid_test_ipv6.sock'

    # remove below one sleep will fail
    # seems like connection_made/data_received only run one
    # but why INET/INET6 does not has this issue?
    await sleep(0)
    await stream.write(b'from pymaid')
    # skip one step to let data_received run
    await sleep(0)

    assert server.connected_stream
    assert server.connected_stream.received_data == b'from pymaid'

    stream.close()
    client.close()
    server.connected_stream.close()
    server.close()
    os.unlink('/tmp/pymaid_test_ipv6.sock')



from typing import Sequence

from pymaid.utils.logger import logger_wrapper

from .types import RouterType, ServiceType


class Router:

    def __init__(
        self,
        *,
        services: Sequence[ServiceType] = [],
        routers: Sequence[RouterType] = [],
    ):
        self.routes = {}
        for service in services:
            self.include_service(service)
        for router in routers:
            self.include_router(router)

    def include_router(self, router: RouterType):
        self.routes.update(router.routes)

    def include_service(self, service: ServiceType):
        routes = self.routes
        for method in self.get_service_methods(service):
            assert method.full_name not in routes
            routes[method.full_name] = method
            # js/lua pb lib will format as '.service.method'
            routes['.' + method.full_name] = method

    def include_services(self, services: Sequence[ServiceType]):
        for service in services:
            self.include_service(service)

    def get_service_methods(self, service: ServiceType):
        pass

    def get_route(self, name):
        return self.routes.get(name)


@logger_wrapper
class RouterStub:

    def __init__(self, stub):
        self.stub = stub
        self.routes = {}
        self.build_method_stub()

    def build_method_stub(self):
        routes = self.routes
        for method_stub in self.get_router_stubs(self.stub):
            setattr(self, method_stub.name, method_stub)
            routes[method_stub.full_name] = method_stub

    def get_router_stubs(self):
        raise NotImplementedError('get_router_stubs')



import struct
from typing import Optional, Sequence, Tuple, TypeVar

from google.protobuf.message import Message

from pymaid.net.protocol import DataType, Protocol

from .error import PBError
from .pymaid_pb2 import Context as Meta


Message = TypeVar('Message', bound=Message)


class Protocol(Protocol):

    HEADER_FORMAT = '!HH'
    HEADER_STRUCT = struct.Struct(HEADER_FORMAT)

    MAX_PACKET_LENGTH = 8 * 1024

    header_size = HEADER_STRUCT.size
    pack_header = HEADER_STRUCT.pack
    unpack_header = HEADER_STRUCT.unpack

    @classmethod
    def feed_data(cls, data: DataType) -> Tuple[int, Sequence[Message]]:
        data = memoryview(data)
        messages = []

        used_size = 0
        try:
            while 1:
                consumed, meta, payload = cls.decode(data)
                if not consumed:
                    break
                messages.append((meta, payload))
                used_size += consumed
                data = data[consumed:]
        finally:
            data.release()

        return used_size, messages

    @classmethod
    def encode(cls, meta: Meta, message: Message) -> bytes:
        return (
            cls.pack_header(meta.ByteSize(), message.ByteSize())
            + meta.SerializeToString()
            + message.SerializeToString()
        )

    @classmethod
    def decode(
        cls, data: DataType,
    ) -> Tuple[int, Optional[Meta], Optional[memoryview]]:
        header_size = cls.header_size
        if (data_size := len(data)) < header_size:
            return 0, None, None

        meta_size, payload_size = cls.unpack_header(data[:header_size])
        if (used_size := header_size + meta_size + payload_size) > data_size:
            return 0, None, None
        if payload_size > cls.MAX_PACKET_LENGTH:
            raise PBError.PacketTooLarge(
                data={'max': cls.MAX_PACKET_LENGTH, 'size': payload_size}
            )

        return (
            used_size,
            Meta.FromString(data[header_size:header_size + meta_size]),
            data[header_size + meta_size: used_size],
        )



from orjson import dumps

from pymaid.rpc.handler import Handler, SerialHandler, ParallelHandler
from pymaid.utils.logger import logger_wrapper

from .context import PBInboundContext, PBOutboundContext
from .error import PBError
from .pymaid_pb2 import Context as Meta, ErrorMessage


@logger_wrapper
class PBHandler(Handler):

    INBOUND_CONTEXT_CLASS = PBInboundContext
    OUTBOUND_CONTEXT_CLASS = PBOutboundContext

    def feed_messages(self, messages):
        Request = Meta.PacketType.REQUEST
        Response = Meta.PacketType.RESPONSE
        get_route = self.router.get_route
        for message in messages:
            # check exist context here for a shortcut
            # because just feed message into context won't block,
            # and it makes serial streaming posible
            meta, payload = message
            # self.logger.debug(f'{self} feed {meta=}')
            if meta.transmission_id in self.contexts:
                self.contexts[meta.transmission_id].feed_message(meta, payload)
                continue

            if meta.packet_type == Request:
                name = meta.service_method
                if (rpc := get_route(name)) is None:
                    task = self.handle_error(
                        meta, PBError.RPCNotFound(data={'name': name})
                    )
                else:
                    context = self.new_inbound_context(
                        meta.transmission_id, method=rpc, timeout=self.timeout
                    )
                    context.feed_message(meta, payload)
                    task = context.run()
            elif meta.packet_type == Response:
                # response should be handled above as existed context
                self.logger.warning(
                    f'{self!r} received unknown response, '
                    f'{meta.transmission_id=}, ignored.'
                )
                continue
            else:
                task = self.handle_error(
                    meta,
                    PBError.InvalidPacketType(
                        data={'packet_type': meta.packet_type}
                    )
                )
            self.pending_tasks.append(task)
            self.new_task_received.set()

    async def handle_error(self, meta, error):
        meta.is_failed = True
        meta.packet_type = Meta.PacketType.RESPONSE
        packet = ErrorMessage(code=error.code, message=error.message)
        if error.data:
            packet.data = dumps(error.data)
        await self.conn.send_message(meta, packet)


@logger_wrapper
class PBSerialHandler(PBHandler, SerialHandler):
    pass


@logger_wrapper
class PBParallelHandler(PBHandler, ParallelHandler):
    pass



# inspired by https://github.com/aio-libs/async-timeout

import os
import time

import pytest

import pymaid
from pymaid.utils.timeout import timeout, timeout_at


@pytest.mark.asyncio
async def test_timeout():
    canceled_raised = False

    async def long_running_task():
        try:
            await pymaid.sleep(10)
        except pymaid.CancelledError:
            nonlocal canceled_raised
            canceled_raised = True
            raise

    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.01) as t:
            await long_running_task()
            assert t._loop is pymaid.get_event_loop()
    assert canceled_raised, 'CancelledError was not raised'


@pytest.mark.asyncio
async def test_timeout_finish_in_time():
    async def long_running_task():
        await pymaid.sleep(0.01)
        return 'done'

    async with timeout(0.1):
        resp = await long_running_task()
    assert resp == 'done'


@pytest.mark.asyncio
async def test_timeout_disable():
    async def long_running_task():
        await pymaid.sleep(0.1)
        return 'done'

    loop = pymaid.get_event_loop()
    t0 = loop.time()
    async with timeout(None):
        resp = await long_running_task()
    assert resp == 'done'
    dt = loop.time() - t0
    assert 0.09 < dt < 0.13, dt


@pytest.mark.asyncio
async def test_timeout_is_none_no_schedule():
    async with timeout(None) as cm:
        assert cm.timeout_handler is None
        assert cm.deadline is None


def test_timeout_no_loop():
    with pytest.raises(RuntimeError, match='no running event loop'):
        timeout(None)


@pytest.mark.asyncio
async def test_timeout_zero():
    with pytest.raises(pymaid.TimeoutError):
        timeout(0)


@pytest.mark.asyncio
async def test_timeout_not_relevant_exception():
    await pymaid.sleep(0)
    with pytest.raises(KeyError):
        async with timeout(0.1):
            raise KeyError


@pytest.mark.asyncio
async def test_timeout_cancelled_error_is_not_converted_to_timeout():
    await pymaid.sleep(0)
    with pytest.raises(pymaid.CancelledError):
        async with timeout(0.001):
            raise pymaid.CancelledError


@pytest.mark.asyncio
async def test_timeout_blocking_loop():
    async def long_running_task():
        time.sleep(0.1)
        return 'done'

    async with timeout(0.01):
        result = await long_running_task()
    assert result == 'done'


@pytest.mark.asyncio
async def test_for_race_conditions():
    loop = pymaid.get_event_loop()
    fut = loop.create_future()
    loop.call_later(0.1, fut.set_result('done'))
    async with timeout(0.2):
        resp = await fut
    assert resp == 'done'


@pytest.mark.asyncio
async def test_timeout_time():
    foo_running = None
    loop = pymaid.get_event_loop()
    start = loop.time()
    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.1):
            foo_running = True
            try:
                await pymaid.sleep(0.2)
            finally:
                foo_running = False

    dt = loop.time() - start
    if not (0.09 < dt < 0.11) and os.environ.get('APPVEYOR'):
        pytest.xfail('appveyor sometimes is toooo sloooow')
    assert 0.09 < dt < 0.11
    assert not foo_running


@pytest.mark.asyncio
async def test_outer_coro_is_not_cancelled():

    has_timeout = False

    async def outer():
        nonlocal has_timeout
        try:
            async with timeout(0.001):
                await pymaid.sleep(1)
        except pymaid.TimeoutError:
            has_timeout = True

    task = pymaid.ensure_future(outer())
    await task
    assert has_timeout
    assert not task.cancelled()
    assert task.done()


@pytest.mark.asyncio
async def test_cancel_outer_coro():
    loop = pymaid.get_event_loop()
    fut = loop.create_future()

    async def outer():
        fut.set_result(None)
        await pymaid.sleep(1)

    task = pymaid.ensure_future(outer())
    await fut
    task.cancel()
    with pytest.raises(pymaid.CancelledError):
        await task
    assert task.cancelled()
    assert task.done()


@pytest.mark.asyncio
async def test_timeout_suppress_exception_chain():
    with pytest.raises(pymaid.TimeoutError) as ctx:
        async with timeout(0.01):
            await pymaid.sleep(10)
    assert not ctx.value.__suppress_context__


@pytest.mark.asyncio
async def test_timeout_expired():
    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.01) as cm:
            await pymaid.sleep(10)
    assert cm.expired


@pytest.mark.asyncio
async def test_timeout_inner_timeout_error():
    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.01) as cm:
            raise pymaid.TimeoutError
    assert not cm.expired


@pytest.mark.asyncio
async def test_timeout_inner_other_error():
    class MyError(RuntimeError):
        pass

    with pytest.raises(MyError):
        async with timeout(0.01) as cm:
            raise MyError
    assert not cm.expired


@pytest.mark.asyncio
async def test_timeout_at():
    loop = pymaid.get_event_loop()
    with pytest.raises(pymaid.TimeoutError):
        now = loop.time()
        async with timeout_at(now + 0.01) as cm:
            await pymaid.sleep(10)
    assert cm.expired


@pytest.mark.asyncio
async def test_timeout_at_not_fired():
    loop = pymaid.get_event_loop()
    now = loop.time()
    async with timeout_at(now + 1) as cm:
        await pymaid.sleep(0)
    assert not cm.expired


@pytest.mark.asyncio
async def test_expired_after_aborting():
    t = timeout(10)
    assert not t.expired
    t.abort()
    assert not t.expired


@pytest.mark.asyncio
async def test_abort_finished():
    async with timeout(10) as t:
        await pymaid.sleep(0)

    assert not t.expired
    with pytest.raises(RuntimeError, match='invalid state EXIT'):
        t.abort()


@pytest.mark.asyncio
async def test_expired_after_timeout():
    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.01) as t:
            assert not t.expired
            await pymaid.sleep(10)
    assert t.expired


@pytest.mark.asyncio
async def test_deadline():
    loop = pymaid.get_event_loop()
    t0 = loop.time()
    async with timeout(1) as cm:
        t1 = loop.time()
        assert t0 + 1 <= cm.deadline <= t1 + 1


@pytest.mark.asyncio
async def test_async_timeout():
    with pytest.raises(pymaid.TimeoutError):
        async with timeout(0.01) as cm:
            await pymaid.sleep(10)
    assert cm.expired


@pytest.mark.asyncio
async def test_async_no_timeout():
    async with timeout(1) as cm:
        await pymaid.sleep(0)
    assert not cm.expired


@pytest.mark.asyncio
async def test_shift_to():
    loop = pymaid.get_event_loop()
    t0 = loop.time()
    async with timeout(1) as cm:
        t1 = loop.time()
        assert t0 + 1 <= cm.deadline <= t1 + 1
        cm.shift_to(t1 + 1)
        assert t1 + 1 <= cm.deadline <= t1 + 1.001


@pytest.mark.asyncio
async def test_shift_by():
    loop = pymaid.get_event_loop()
    t0 = loop.time()
    async with timeout(1) as cm:
        t1 = loop.time()
        assert t0 + 1 <= cm.deadline <= t1 + 1
        cm.shift_by(1)
        assert t1 + 0.999 <= cm.deadline <= t1 + 1.001


@pytest.mark.asyncio
async def test_shift_by_negative_expired():
    async with timeout(1) as cm:
        with pytest.raises(pymaid.CancelledError):
            cm.shift_by(-1)


@pytest.mark.asyncio
async def test_shift_by_expired():
    async with timeout(0.001) as cm:
        with pytest.raises(pymaid.CancelledError):
            await pymaid.sleep(10)
        with pytest.raises(RuntimeError, match='cannot reschedule expired timeout'):
            await cm.shift_by(10)


@pytest.mark.asyncio
async def test_shift_to_expired():
    loop = pymaid.get_event_loop()
    t0 = loop.time()
    async with timeout_at(t0 + 0.001) as cm:
        with pytest.raises(pymaid.CancelledError):
            await pymaid.sleep(10)
        with pytest.raises(RuntimeError, match='cannot reschedule expired timeout'):
            await cm.shift_to(t0 + 10)


@pytest.mark.asyncio
async def test_shift_by_after_cm_exit():
    async with timeout(1) as cm:
        await pymaid.sleep(0)
    with pytest.raises(
        RuntimeError, match='cannot reschedule after exit from context manager'
    ):
        cm.shift_by(1)


@pytest.mark.asyncio
async def test_enter_twice():
    async with timeout(10) as t:
        await pymaid.sleep(0)

    with pytest.raises(RuntimeError, match='invalid state EXIT'):
        async with t:
            await pymaid.sleep(0)



from typing import TypeVar

from pymaid.net.stream import Stream


class Connection(Stream):
    '''Connection represent a communication way for client <--> server

    It holds the low level transport.
    '''

    KEEP_OPEN_ON_EOF = True

    def __init__(
        self,
        sock,
        *,
        protocol,
        handler,
        **kwargs
    ):
        super().__init__(sock, **kwargs)
        self.protocol = protocol
        self.handler = handler
        self.read_buffer = bytearray()

        # NOTE: cyclic
        handler.start(self)

    def data_received(self, data: bytes):
        '''Received data from low level transport'''
        self.read_buffer.extend(data)
        used_size, messages = self.protocol.feed_data(self.read_buffer)
        if used_size:
            self.read_buffer = self.read_buffer[used_size:]
            self.handler.feed_messages(messages)

    def eof_received(self):
        self.handler.shutdown('eof_received')
        return super().eof_received()

    async def send_message(self, *args, **kwargs):
        '''Helper to send protocol message'''
        await self.write(self.protocol.encode(*args, **kwargs))

    def close(self, exc=None):
        if self.state == self.STATE.CLOSED:
            return
        self.handler.shutdown(exc)
        super().close(exc)

    def _finnal_close(self, exc=None):
        super()._finnal_close(exc)
        self.handler.close(exc)
        del self.handler
        del self.read_buffer[:]


ConnectionType = TypeVar('Connection', bound=Connection)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymaid/rpc/pb/pymaid.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name='pymaid/rpc/pb/pymaid.proto',
    package='pymaid.rpc.pb',
    syntax='proto3',
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1apymaid/rpc/pb/pymaid.proto\x12\rpymaid.rpc.pb\"\xb1\x03\n\x07\x43ontext\x12\x17\n\x0ftransmission_id\x18\x01 \x01(\r\x12\x36\n\x0bpacket_type\x18\x02 \x01(\x0e\x32!.pymaid.rpc.pb.Context.PacketType\x12\x37\n\x0cpacket_flags\x18\x03 \x01(\x0e\x32!.pymaid.rpc.pb.Context.PacketFlag\x12\x31\n\x08priority\x18\x04 \x01(\x0e\x32\x1f.pymaid.rpc.pb.Context.Priority\x12\x16\n\x0eservice_method\x18\x05 \x01(\t\x12\x14\n\x0cis_cancelled\x18\x06 \x01(\x08\x12\x11\n\tis_failed\x18\x07 \x01(\x08\x12\x14\n\x0cpayload_size\x18\x08 \x01(\r\"4\n\nPacketType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07REQUEST\x10\x01\x12\x0c\n\x08RESPONSE\x10\x02\"4\n\nPacketFlag\x12\x08\n\x04NULL\x10\x00\x12\x07\n\x03NEW\x10\x01\x12\n\n\x06\x43\x41NCEL\x10\x02\x12\x07\n\x03\x45ND\x10\x04\"&\n\x08Priority\x12\x07\n\x03LOW\x10\x00\x12\x07\n\x03MID\x10\x01\x12\x08\n\x04HIGH\x10\x02\"\x08\n\x06RpcAck\"\x06\n\x04Void\";\n\x0c\x45rrorMessage\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\tb\x06proto3'
)


_CONTEXT_PACKETTYPE = _descriptor.EnumDescriptor(
    name='PacketType',
    full_name='pymaid.rpc.pb.Context.PacketType',
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name='UNKNOWN', index=0, number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='REQUEST', index=1, number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='RESPONSE', index=2, number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=333,
    serialized_end=385,
)
_sym_db.RegisterEnumDescriptor(_CONTEXT_PACKETTYPE)

_CONTEXT_PACKETFLAG = _descriptor.EnumDescriptor(
    name='PacketFlag',
    full_name='pymaid.rpc.pb.Context.PacketFlag',
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name='NULL', index=0, number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='NEW', index=1, number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='CANCEL', index=2, number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='END', index=3, number=4,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=387,
    serialized_end=439,
)
_sym_db.RegisterEnumDescriptor(_CONTEXT_PACKETFLAG)

_CONTEXT_PRIORITY = _descriptor.EnumDescriptor(
    name='Priority',
    full_name='pymaid.rpc.pb.Context.Priority',
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name='LOW', index=0, number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='MID', index=1, number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='HIGH', index=2, number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=441,
    serialized_end=479,
)
_sym_db.RegisterEnumDescriptor(_CONTEXT_PRIORITY)


_CONTEXT = _descriptor.Descriptor(
    name='Context',
    full_name='pymaid.rpc.pb.Context',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='transmission_id', full_name='pymaid.rpc.pb.Context.transmission_id', index=0,
            number=1, type=13, cpp_type=3, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='packet_type', full_name='pymaid.rpc.pb.Context.packet_type', index=1,
            number=2, type=14, cpp_type=8, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='packet_flags', full_name='pymaid.rpc.pb.Context.packet_flags', index=2,
            number=3, type=14, cpp_type=8, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='priority', full_name='pymaid.rpc.pb.Context.priority', index=3,
            number=4, type=14, cpp_type=8, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='service_method', full_name='pymaid.rpc.pb.Context.service_method', index=4,
            number=5, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='is_cancelled', full_name='pymaid.rpc.pb.Context.is_cancelled', index=5,
            number=6, type=8, cpp_type=7, label=1,
            has_default_value=False, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='is_failed', full_name='pymaid.rpc.pb.Context.is_failed', index=6,
            number=7, type=8, cpp_type=7, label=1,
            has_default_value=False, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='payload_size', full_name='pymaid.rpc.pb.Context.payload_size', index=7,
            number=8, type=13, cpp_type=3, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
        _CONTEXT_PACKETTYPE,
        _CONTEXT_PACKETFLAG,
        _CONTEXT_PRIORITY,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=46,
    serialized_end=479,
)


_RPCACK = _descriptor.Descriptor(
    name='RpcAck',
    full_name='pymaid.rpc.pb.RpcAck',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=481,
    serialized_end=489,
)


_VOID = _descriptor.Descriptor(
    name='Void',
    full_name='pymaid.rpc.pb.Void',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=491,
    serialized_end=497,
)


_ERRORMESSAGE = _descriptor.Descriptor(
    name='ErrorMessage',
    full_name='pymaid.rpc.pb.ErrorMessage',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='code', full_name='pymaid.rpc.pb.ErrorMessage.code', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='message', full_name='pymaid.rpc.pb.ErrorMessage.message', index=1,
            number=2, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='data', full_name='pymaid.rpc.pb.ErrorMessage.data', index=2,
            number=3, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR, create_key=_descriptor._internal_create_key),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=499,
    serialized_end=558,
)

_CONTEXT.fields_by_name['packet_type'].enum_type = _CONTEXT_PACKETTYPE
_CONTEXT.fields_by_name['packet_flags'].enum_type = _CONTEXT_PACKETFLAG
_CONTEXT.fields_by_name['priority'].enum_type = _CONTEXT_PRIORITY
_CONTEXT_PACKETTYPE.containing_type = _CONTEXT
_CONTEXT_PACKETFLAG.containing_type = _CONTEXT
_CONTEXT_PRIORITY.containing_type = _CONTEXT
DESCRIPTOR.message_types_by_name['Context'] = _CONTEXT
DESCRIPTOR.message_types_by_name['RpcAck'] = _RPCACK
DESCRIPTOR.message_types_by_name['Void'] = _VOID
DESCRIPTOR.message_types_by_name['ErrorMessage'] = _ERRORMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Context = _reflection.GeneratedProtocolMessageType('Context', (_message.Message,), {
    'DESCRIPTOR': _CONTEXT,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.Context)
})
_sym_db.RegisterMessage(Context)

RpcAck = _reflection.GeneratedProtocolMessageType('RpcAck', (_message.Message,), {
    'DESCRIPTOR': _RPCACK,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.RpcAck)
})
_sym_db.RegisterMessage(RpcAck)

Void = _reflection.GeneratedProtocolMessageType('Void', (_message.Message,), {
    'DESCRIPTOR': _VOID,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.Void)
})
_sym_db.RegisterMessage(Void)

ErrorMessage = _reflection.GeneratedProtocolMessageType('ErrorMessage', (_message.Message,), {
    'DESCRIPTOR': _ERRORMESSAGE,
    '__module__': 'pymaid.rpc.pb.pymaid_pb2'
    # @@protoc_insertion_point(class_scope:pymaid.rpc.pb.ErrorMessage)
})
_sym_db.RegisterMessage(ErrorMessage)


# @@protoc_insertion_point(module_scope)



from collections import defaultdict
from typing import List, Optional


class BaseMiddleware:
    '''Base class for Middleware.

    *NOTE*: all on_* methods will be treated as event callbacks,
        and will be cached in MiddlewareManager.events for faster dispatching.

    e.g.: on_connect/on_close are for examples.
    '''

    def on_start(self, *args, **kwargs):
        pass

    def on_pause(self, *args, **kwargs):
        pass

    def on_shutdown(self, *args, **kwargs):
        pass

    def on_close(self, *args, **kwargs):
        pass


class MiddlewareManager:

    def __init__(self, middlewares: Optional[List[BaseMiddleware]] = None):
        self.middlewares = []
        self.events = defaultdict(list)

        if middlewares:
            self.append_middlewares(middlewares)

    def append_middleware(self, middleware: BaseMiddleware):
        self.middlewares.append(middleware)

        for attr in dir(middleware):
            if attr.startswith('on_'):
                self.events[attr].append(getattr(middleware, attr))

    def append_middlewares(self, middlewares: List[BaseMiddleware]):
        for mw in middlewares:
            self.append_middleware(mw)

    def dispatch(self, event, *args, **kwargs):
        for callback in self.events.get(event, []):
            callback(*args, **kwargs)



import os
from unittest import TestCase

from pymaid.conf import defaults, settings


class SettingsTest(TestCase):

    def setUp(self):
        global settings
        settings.namespaces.clear()
        settings.load_from_object(defaults, ns='pymaid')

        self.other_namespace = {
            '__NAMESPACE__': 'other',
            '__MUTABLE__': False,
            'key': 1,
            'KEY': 2,
        }

    def test_get(self):
        assert settings.get('DEBUG', ns='pymaid') == defaults.DEBUG
        assert settings.get('NOT_EXIST', default='what') == 'what'

        settings.load_from_object(self.other_namespace)
        assert settings.get('KEY', ns='other') == self.other_namespace['KEY']
        # only upper key will be treated as settings
        assert settings.get('key', ns='other') is None

    def test_set(self):
        assert settings.get('MAX_TASKS', ns='pymaid') == defaults.MAX_TASKS
        settings.set('MAX_TASKS', defaults.MAX_TASKS + 1, ns='pymaid')
        assert settings.get('MAX_TASKS', ns='pymaid') != defaults.MAX_TASKS
        assert settings.get('MAX_TASKS', ns='pymaid') == defaults.MAX_TASKS + 1

        settings.load_from_object(self.other_namespace)
        assert settings.get('KEY', ns='other') == self.other_namespace['KEY']
        # cannot set immutable settings
        with self.assertRaises(RuntimeError):
            settings.set('KEY', 3, ns='other')

        # cannot set not exist settings
        with self.assertRaises(AttributeError):
            settings.set('NOT_EXIST', 1, ns='pymaid')

    def test_load_env(self):
        assert settings.get('MAX_TASKS', ns='pymaid') == defaults.MAX_TASKS
        assert settings.get('NOT_EXIST', default='what') == 'what'

        os.environ['SETTING__PYMAID__NOT_EXIST'] = 'str::noway'
        os.environ['SETTING__COMMON__NOT_EXIST'] = 'bytes::noway'
        os.environ['SETTING__PYMAID__MAX_TASKS'] = 'int::128'
        os.environ['SETTING__PYMAID__FLOAT'] = 'float::5.13543512414351235'
        os.environ['SETTING__PYMAID__BOOL'] = 'bool::False'
        os.environ['SETTING__PYMAID__DICT'] = 'dict::{"age": 18}'
        os.environ['SETTING__PYMAID__LIST'] = 'list::[1, 2, 3]'
        settings.load_from_environment(prefix='SETTING__')

        assert settings.get('NOT_EXIST', ns='pymaid') == 'noway'
        assert settings.get('NOT_EXIST', default='what') == b'noway'
        assert settings.get('MAX_TASKS', ns='pymaid') == 128
        assert abs(settings.get('FLOAT', ns='pymaid') - 5.13543512414351235) < 1e-6  # noqa
        assert settings.get('BOOL', ns='pymaid') is False
        assert settings.get('DICT', ns='pymaid') == {'age': 18}
        assert settings.get('LIST', ns='pymaid') == [1, 2, 3]

    def test_load_env_bad_cases(self):
        # wrong prefix
        os.environ.clear()
        settings.namespaces.clear()

        os.environ['SETTING__PYMAID__LIST'] = 'list::[1, 2, 3]'
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING_')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        os.environ.clear()
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING___')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        # wrong namespace
        os.environ['SETTING___PYMAID__LIST'] = 'list::[1, 2, 3]'
        settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        # wrong key
        os.environ['SETTING__PYMAID___LIST'] = 'list::[1, 2, 3]'
        settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        # wrong value format
        os.environ['SETTING__PYMAID__LIST'] = '[1, 2, 3]'
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        os.environ['SETTING__PYMAID__LIST'] = 'list:[1, 2, 3]'
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        # wrong value type
        os.environ['SETTING__PYMAID__LIST'] = 'what_type::[1, 2, 3]'
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

        # wrong type::value
        os.environ['SETTING__PYMAID__LIST'] = 'int::[1, 2, 3]'
        with self.assertRaises(ValueError):
            settings.load_from_environment(prefix='SETTING__')
        assert settings.get('LIST', ns='pymaid') is None
        assert not settings.namespaces

    def test_dot_get(self):
        assert settings.pymaid.DEBUG == defaults.DEBUG

        # not exists namespace
        with self.assertRaises(AttributeError):
            settings.what

        # not exists key
        with self.assertRaises(AttributeError):
            settings.pymaid.what

    def test_dot_set(self):
        assert settings.pymaid.MAX_TASKS == defaults.MAX_TASKS
        settings.pymaid.MAX_TASKS += 1
        assert settings.pymaid.MAX_TASKS == defaults.MAX_TASKS + 1

        # cannot set not exist
        with self.assertRaises(AttributeError):
            settings.pymaid.NOT_EXIST = 1

        settings.load_from_object(self.other_namespace)
        # cannot set immutable
        with self.assertRaises(RuntimeError):
            settings.other.KEY = 1



from __future__ import absolute_import

from collections import defaultdict
from functools import partial
from importlib import import_module
import os
import re
import sys

from orjson import loads

from pymaid.utils.logger import configure_logging, logger_wrapper

from . import defaults


class Namespace(dict):

    def __dir__(self):
        return super(Namespace, self).__dir__() + list(self.keys())

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if not self.get('__MUTABLE__', False):
            raise RuntimeError('settings namespace is immutable')
        if name in self:
            self[name] = value
        else:
            raise AttributeError(name)


@logger_wrapper
class Settings(object):

    transformer = {
        'str': str,
        'bytes': partial(bytes, encoding='utf-8'),
        'int': partial(int, base=10),
        'float': float,
        'bool': lambda x: False if x in ('False', 'false') else True,
        'dict': loads,
        'list': loads,
    }

    def __init__(self, name):
        self.name = name
        self.namespaces = {}
        self.watchers = defaultdict(list)

    def filter(self, key, value):
        return key == key.upper()

    def get(self, key, default=None, ns='common'):
        '''Return specified config value in local namespaces cache.

        This method will not block and return default value if key not exists.
        User should initiate settings manually by calling `load_from_*` apis.
        '''
        try:
            return self.namespaces[ns][key]
        except KeyError:
            return default

    def set(self, key, value, ns='common'):
        '''Update specified config value in local namespaces cache.

        This method will not block and *will not* update remote settings,
        this change will be discarded after destory this process.

        Fix: check __MUTABLE__ before setting settings
        '''
        try:
            namespace = self.namespaces[ns]
        except KeyError:
            self.logger.warn(
                'set value: %s to key: %s, of unknown namespace: %s',
                value, key, ns,
            )
        if not namespace.get('__MUTABLE__', False):
            raise RuntimeError(f'[pymaid][settings][ns|{ns}] is immutable')
        setattr(namespace, key, value)

    def add_watcher(self, watcher, ns='common'):
        if watcher in self.watchers[ns]:
            return
        self.watchers[ns].append(watcher)

    def load_from_object(self, obj, ns='common', mutable=True):
        if isinstance(obj, dict):
            ns = obj.get('__NAMESPACE__', ns)
        else:
            ns = getattr(obj, '__NAMESPACE__', ns)

        if (ns in self.namespaces
                and not self.namespaces[ns].get('__MUTABLE__', mutable)):
            self.logger.warn('[pymaid][settings][ns|%s] is immutable', ns)
            return False

        data = {}
        filter = self.filter
        if isinstance(obj, dict):
            data.update({
                key: value for key, value in obj.items() if filter(key, value)
            })
        else:
            data.update({
                key: getattr(obj, key)
                for key in dir(obj) if filter(key, getattr(obj, key))
            })
        self.namespaces.setdefault(ns, Namespace()).update(data)
        self.namespaces[ns].setdefault(
            '__MUTABLE__', data.get('__MUTABLE__', mutable)
        )

        for watcher in self.watchers[ns]:
            watcher(self, ns)
        return True

    def load_from_module(self, module_name, ns='common', mutable=True):
        '''Load the settings module pointed to by the module_name.

        This is used the first time we need any settings at all,
        if the user has not previously configured the settings manually.

        The user can manually configure settings prior to using them.
        '''
        try:
            mod = import_module(module_name)
        except ImportError as ex:
            raise ImportError(
                f'Could not import settings `{module_name}` '
                f'(Is it on sys.path?): {ex}'
            )
        self.load_from_object(mod, getattr(mod, '__NAMESPACE__', ns), mutable)

    def load_from_root_path(self, path, ns='common', mutable=True):
        for root, dirs, files in os.walk(path):
            if '__init__.py' not in files:
                continue
            try:
                # import settings explicitly
                mod = import_module(root.replace('/', '.') + '.settings')
            except ImportError as ex:
                if re.search('No module named .*settings', str(ex)):
                    continue
                raise
            else:
                self.load_from_object(
                    mod, getattr(mod, '__NAMESPACE__', ns), mutable
                )

    def load_from_environment(
        self, prefix='SETTING__', raise_invalid_value=True
    ):
        '''load *special formatted* env into settings

        format: {PREFIX}__{NAMESPACE}__{KEY}=VALUE
        PREFIX: pymaid settings
        NAMESPACE: namespace
        KEY: settings key name
        VALUE: type::value, type need to be builtin type,
            current are %s
            dict/list will be loaded using json.loads

        NOTE: when loaded, {NAMESPACE} will transform to lower case

        e.g.:
            export SETTING__PYMAID__DEBUG='bool::True'
            will result in below
            settings.set('DEBUG', True, ns='pymaid')
            settings.namespaces['pymaid']['DEBUG'] = True
        '''

        # cannot endswith ___
        # can only has one '__'
        if not re.match(r'[A-Z]+__$', prefix):
            raise ValueError(
                f'prefix should be in the format of `NAME__`, got `{prefix}`'
            )

        env_regex = re.compile(
            r'^%s[A-Z][A-Z_]+[A-Z]__[A-Z][A-Z_]+[A-Z]$' % prefix
        )
        data = {}
        for env, value in os.environ.items():
            # naive check
            if not env.startswith(prefix):
                continue
            if not env_regex.match(env):
                sys.stderr.write(f'wrong special formatted env `{env}`\n')
                continue

            env_ = env.split('__')
            if len(env_) != 3:
                sys.stderr.write(f'wrong special formatted env `{env}`\n')
                continue
            _, ns, key = env_

            value_ = value.split('::')
            if len(value_) != 2:
                err = (
                    f'get special formatted env `{env}`, '
                    f'but with wrong format value `{value}`, '
                    f'should be in format of `type::value`\n'
                )
                if raise_invalid_value:
                    raise ValueError(err)
                sys.stderr.write(err)
                continue

            t, val = value_
            if t not in self.transformer:
                err = (
                    f'unknown value type `{t}` for env `{env}={value}`, '
                    f'available `{self.transformer.keys()}`\n'
                )
                if raise_invalid_value:
                    raise ValueError(err)
                sys.stderr.write(err)
                continue

            try:
                val = self.transformer[t](val)
            except (TypeError, ValueError):
                err = (
                    f'cannot transform value of `{env}={value}`, '
                    f'check type and value\n'
                )
                if raise_invalid_value:
                    raise ValueError(err)
                sys.stderr.write(err)
                continue

            # now we successfully get special env=value
            ns_data = data.setdefault(ns.lower(), {})
            ns_data[key] = val
        sys.stderr.flush()

        for ns, obj in data.items():
            self.load_from_object(obj, ns=ns)

    load_from_environment.__doc__ = (
        load_from_environment.__doc__ % list(transformer.keys())
    )

    def __str__(self):
        return f'[{self.name}][namespaces|{len(self.namespaces)}]'
    __repr__ = __str__

    def __getattr__(self, name):
        '''Implemented `settings.namespace.key` usage'''
        if name in self.namespaces:
            # cache
            ns = self.namespaces[name]
            setattr(self, name, ns)
            return ns
        raise AttributeError(name)

    def __dir__(self):
        return super(Settings, self).__dir__() + list(self.namespaces.keys())


settings = Settings('global')
settings.add_watcher(configure_logging, ns='pymaid')
settings.add_watcher(configure_logging, ns='logging')
settings.load_from_object(defaults, ns='pymaid')
settings.load_from_environment()



import struct

from itertools import cycle
from typing import Tuple
from urllib.parse import urlparse

from pymaid.conf import settings
from pymaid.net.protocol import Protocol
from pymaid.types import DataType

from .exceptions import ProtocolError, FrameTooLargeException

try:
    from .speedups import apply_mask
    mask_payload = unmask_payload = apply_mask
except ImportError:

    # from_bytes/to_bytes is faster
    # but it is under potential risk of being attack
    # because multiply *mask* is not memory friendly

    # def mask_payload(payload, mask) -> bytes:
    #     p_size = len(payload)
    #     m_size = len(mask)
    #     if p_size > m_size:
    #         q = p_size // m_size
    #         r = p_size % m_size
    #         mask = mask * q + mask[:r]
    #     elif p_size < m_size:
    #         mask = mask[:p_size]

    #     return (
    #         int.from_bytes(payload, 'little') ^ int.from_bytes(mask, 'little')  # noqa
    #     ).to_bytes(p_size, 'little')

    def mask_payload(payload: DataType, mask: bytes) -> bytes:
        if len(mask) != 4:
            raise ValueError('mask must be 4 bytes')
        return bytes(b ^ m for b, m in zip(payload, cycle(mask)))

    # it's the same operation
    unmask_payload = mask_payload


__all__ = ['WSProtocol', 'Header', 'mask_payload', 'unmask_payload']

st = struct.Struct('!BB')
pack_header = st.pack
unpack_header = st.unpack

st_H = struct.Struct('!H')
pack_H = st_H.pack
unpack_H = st_H.unpack

st_Q = struct.Struct('!Q')
pack_Q = st_Q.pack
unpack_Q = st_Q.unpack


def parse_url(url: str):
    if ':' not in url:
        raise ValueError('url is invalid')

    scheme, url = url.split(':', 1)

    is_secure = False
    port = 0
    if scheme == 'ws':
        if not port:
            port = 80
    elif scheme == 'wss':
        is_secure = True
        if not port:
            port = 443
    else:
        raise ValueError(f'scheme {scheme} is invalid')

    parsed = urlparse(url, scheme='ws')
    if parsed.hostname:
        hostname = parsed.hostname
    else:
        raise ValueError('hostname is invalid')

    if parsed.port:
        port = parsed.port

    if parsed.path:
        resource = parsed.path
    else:
        resource = '/'

    if parsed.query:
        resource += '?' + parsed.query

    return hostname, port, resource, is_secure


class WSProtocol(Protocol):

    HANDSHAKE_REQ = (
        b'GET %s HTTP/1.1\r\n'
        b'Upgrade: websocket\r\n'
        b'Connection: Upgrade\r\n'
        b'Sec-WebSocket-Key: %s\r\n'
        b'Sec-WebSocket-Version: %s\r\n\r\n'
    )
    HANDSHAKE_RESP = (
        b'HTTP/1.1 101 Switching Protocols\r\n'
        b'Upgrade: websocket\r\n'
        b'Connection: Upgrade\r\n'
        b'Sec-WebSocket-Accept: %s\r\n\r\n'
    )

    VERSION = b'13'
    SUPPORTED_VERSIONS = (b'13', b'8', b'7')
    GUID = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

    OPCODE_CONTINUATION = 0x00
    OPCODE_TEXT = 0x01
    OPCODE_BINARY = 0x02
    OPCODE_CLOSE = 0x08
    OPCODE_PING = 0x09
    OPCODE_PONG = 0x0a

    def feed_data(self, data: DataType) -> bytes:
        data = memoryview(data)
        messages = []

        used_size = 0
        max_packet = settings.get('MAX_PACKET_LENGTH', ns='pymaid')
        try:
            while 1:
                consumed, meta, payload = self.decode(data, max_packet)
                if not consumed:
                    break
                messages.append((meta, payload))
                used_size += consumed
                data = data[consumed:]
        finally:
            data.release()

        return used_size, messages

    def encode_frame(self, opcode, payload):
        # no mater what opcode, message should be binary type
        if isinstance(payload, str):
            payload = payload.encode('utf-8')
        return (
            Header.encode_header(True, opcode, b'', len(payload), 0) + payload
        )

    def encode(self, opcode, payload: DataType) -> bytes:
        '''Send a frame over the websocket with message as its payload.'''
        return self.encode_frame(
            self.OPCODE_BINARY
            if isinstance(payload, bytes)
            else self.OPCODE_TEXT,
            payload,
        )

    def decode(self, data: DataType) -> Tuple[int, DataType]:
        pass

    def handshake_frame(self, resource, key):
        return self.HANDSHAKE_REQ % (resource, key, self.VERSION)


class Header:

    __slots__ = ('fin', 'mask', 'opcode', 'flags', 'length')

    FIN_MASK = 0x80
    OPCODE_MASK = 0x0f
    MASK_MASK = 0x80
    LENGTH_MASK = 0x7f

    RSV0_MASK = 0x40
    RSV1_MASK = 0x20
    RSV2_MASK = 0x10

    # bitwise mask that will determine the reserved bits for a frame header
    HEADER_FLAG_MASK = RSV0_MASK | RSV1_MASK | RSV2_MASK

    def __init__(self, fin=0, opcode=0, flags=0, length=0):
        self.mask = bytearray()
        self.fin = fin
        self.opcode = opcode
        self.flags = flags
        self.length = length

    def __repr__(self):
        return (
            f'<Header fin={self.fin} opcode={self.opcode} '
            f'length={self.length} flags={self.flags} at 0x{id(self):x}>'
        )

    @classmethod
    def decode_header(cls, data: DataType) -> Tuple[int, 'Header']:
        '''Decode a WebSocket header.

        :param data: `DataType`
        :returns: A `Header` instance.
        '''
        nbytes = len(data)
        if nbytes < 2:
            return 0, None

        used_size = 2
        header = data[:2]
        first_byte, second_byte = unpack_header(header)
        header = cls(
            fin=first_byte & cls.FIN_MASK == cls.FIN_MASK,
            opcode=first_byte & cls.OPCODE_MASK,
            flags=first_byte & cls.HEADER_FLAG_MASK,
            length=second_byte & cls.LENGTH_MASK
        )
        has_mask = second_byte & cls.MASK_MASK == cls.MASK_MASK

        if header.opcode > 0x07:
            if not header.fin:
                raise ProtocolError(
                    f'Received fragmented control frame: {data}'
                )
            # Control frames MUST have a payload length of 125 bytes or less
            if header.length > 125:
                raise FrameTooLargeException(
                    f'Control frame cannot be larger than 125 bytes: {data}'
                )

        if header.length == 126:
            # 16 bit length
            if nbytes < (used_size + 2):
                return 0, None
            header.length = unpack_H(data[used_size: used_size + 2])[0]
            used_size += 2
        elif header.length == 127:
            # 64 bit length
            if nbytes < (used_size + 8):
                return 0, None
            header.length = unpack_Q(data[used_size: used_size + 8])[0]
            used_size += 8

        if has_mask:
            if nbytes < (used_size + 4):
                return 0, None
            header.mask = bytearray(data[used_size: used_size + 4])
            used_size += 4
        return used_size, header

    @classmethod
    def encode_header(cls, fin, opcode, mask, length, flags) -> bytes:
        '''Encodes a WebSocket header.

        :param fin: Whether this is the final frame for this opcode.
        :param opcode: The opcode of the payload, see `OPCODE_*`
        :param mask: Whether the payload is masked.
        :param length: The length of the frame.
        :param flags: The RSV* flags.
        :return: A bytestring encoded header.
        '''
        first_byte = opcode
        second_byte = 0
        extra = b''

        if fin:
            first_byte |= cls.FIN_MASK
        if flags & cls.RSV0_MASK:
            first_byte |= cls.RSV0_MASK
        if flags & cls.RSV1_MASK:
            first_byte |= cls.RSV1_MASK
        if flags & cls.RSV2_MASK:
            first_byte |= cls.RSV2_MASK

        # now deal with length complexities
        if length < 126:
            second_byte += length
        elif length <= 0xffff:
            second_byte += 126
            extra = pack_H(length)
        elif length <= 0xffffffffffffffff:
            second_byte += 127
            extra = pack_Q(length)
        else:
            raise FrameTooLargeException(length)

        if mask:
            second_byte |= cls.MASK_MASK
            extra += mask
        return pack_header(first_byte, second_byte) + extra



from google.protobuf.descriptor_pb2 import MethodDescriptorProto
from google.protobuf.service_reflection import GeneratedServiceType

from pymaid.rpc.method import UnaryUnaryMethod, UnaryStreamMethod
from pymaid.rpc.method import StreamUnaryMethod, StreamStreamMethod
from pymaid.rpc.method import UnaryUnaryMethodStub, UnaryStreamMethodStub
from pymaid.rpc.method import StreamUnaryMethodStub, StreamStreamMethodStub
from pymaid.rpc.router import Router, RouterStub

from .pymaid_pb2 import Context as Meta, Void


class PBRouter(Router):

    def get_service_methods(self, service: GeneratedServiceType):
        for method in service.DESCRIPTOR.methods:
            method_impl = getattr(service, method.name)
            if '.<lambda>' in str(method_impl):
                continue

            mdp = MethodDescriptorProto()
            method.CopyToProto(mdp)
            if not mdp.client_streaming and not mdp.server_streaming:
                method_class = UnaryUnaryMethod
            elif not mdp.client_streaming and mdp.server_streaming:
                method_class = UnaryStreamMethod
            elif mdp.client_streaming and not mdp.server_streaming:
                method_class = StreamUnaryMethod
            elif mdp.client_streaming and mdp.server_streaming:
                method_class = StreamStreamMethod
            else:
                assert False, 'should be one of above'

            request_class = service.GetRequestClass(method)
            response_class = service.GetResponseClass(method)
            method_ins = method_class(
                method.name,
                method.full_name,
                method_impl,
                request_class,
                response_class,
                options={
                    'flags': Meta.PacketFlag.NULL,
                    'void_request': issubclass(request_class, Void),
                    'void_response': issubclass(response_class, Void),
                },
            )
            yield method_ins


class PBRouterStub(RouterStub):

    def get_router_stubs(self, stub):
        for method in stub.DESCRIPTOR.methods:
            mdp = MethodDescriptorProto()
            method.CopyToProto(mdp)
            if not mdp.client_streaming and not mdp.server_streaming:
                method_class = UnaryUnaryMethodStub
            elif not mdp.client_streaming and mdp.server_streaming:
                method_class = UnaryStreamMethodStub
            elif mdp.client_streaming and not mdp.server_streaming:
                method_class = StreamUnaryMethodStub
            elif mdp.client_streaming and mdp.server_streaming:
                method_class = StreamStreamMethodStub
            else:
                assert False, 'should be one of above'

            request_class = stub.GetRequestClass(method)
            response_class = stub.GetResponseClass(method)
            method_stub = method_class(
                method.name,
                method.full_name,
                request_class,
                response_class,
                options={
                    'flags': Meta.PacketFlag.NULL,
                    'void_request': issubclass(request_class, Void),
                    'void_response': issubclass(response_class, Void),
                },
            )
            yield method_stub



import pymaid
from pymaid.net.stream import Stream

from examples.template import get_client_parser, parse_args


class Stream(Stream):

    def data_received(self, data):
        self.data_size += len(data)


def init(stream):
    stream.data_size = 0


async def wrapper(ch, count):
    stream = await ch.acquire(on_open=[init])

    for _ in range(count):
        await stream.write(b'a' * 8000)
    stream.shutdown()
    await stream.closed_event.wait()
    assert stream.data_size == 8000 * count, (stream.data_size, 8000 * count)


async def main(args):
    ch = await pymaid.net.dial_stream(args.address, transport_class=Stream)
    tasks = []
    for x in range(args.concurrency):
        tasks.append(pymaid.create_task(wrapper(ch, args.request)))

    # await pymaid.wait(tasks, timeout=args.timeout)
    await pymaid.gather(*tasks)


if __name__ == "__main__":
    args = parse_args(get_client_parser())
    pymaid.run(main(args))



import ssl as _ssl
from typing import Optional, Sequence, Tuple, Type, Union

from pymaid.ext.middleware import MiddlewareManager
from pymaid.net.protocol import ProtocolType
from pymaid.rpc.channel import create_stream_channel
from pymaid.rpc.connection import Connection, ConnectionType
from pymaid.rpc.types import ServiceType

from . import context
from . import handler
from . import protocol
from . import router

__all__ = ('connection', 'context', 'handler', 'protocol')


async def dial_stream(
    address: Union[Tuple[str, int], str],
    *,
    name: str = 'StreamChannel',
    transport_class: ConnectionType = Connection,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    close_conn_onerror: bool = True,
    protocol: protocol.Protocol = protocol.Protocol,
    handler_class: Type[handler.Handler] = handler.PBSerialHandler,
    router_class: Type[router.Router] = router.PBRouter,
    middleware_manager: Optional[MiddlewareManager] = None,
):
    return create_stream_channel(
        address,
        name=name,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        close_conn_onerror=close_conn_onerror,
        protocol=protocol,
        handler_class=handler_class,
        router_class=router_class,
        middleware_manager=middleware_manager,
    )


async def serve_stream(
    address: Union[Tuple[str, int], str],
    *,
    name: str = 'StreamChannel',
    transport_class: ConnectionType = Connection,
    backlog: int = 128,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    middleware_manager: Optional[MiddlewareManager] = None,
    close_conn_onerror: bool = True,
    protocol: ProtocolType = protocol.Protocol,
    handler_class: Type[handler.Handler] = handler.PBSerialHandler,
    router_class: Type[router.Router] = router.PBRouter,
    services: Optional[Sequence[ServiceType]] = None,
    router: Optional[router.Router] = None,
):
    assert services is not None or router, 'should provide services or router'
    ch = create_stream_channel(
        address,
        name=name,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        middleware_manager=middleware_manager,
        close_conn_onerror=close_conn_onerror,
        protocol=protocol,
        handler_class=handler_class,
        router_class=router_class,
    )
    if services:
        ch.router.include_services(services)
    if router:
        ch.router.include_router(router)
    await ch.listen(address, backlog=backlog)
    ch.start()
    async with ch:
        await ch.serve_forever()


def implall(service):
    service_name = service.DESCRIPTOR.name
    missing = []
    for base in service.__bases__:
        for method in base.DESCRIPTOR.methods:
            method_name = method.name
            base_method = getattr(base, method_name)
            impl_method = getattr(service, method_name, base_method)
            if base_method == impl_method:
                missing.append(f'{service_name}.{method_name}')
    if missing:
        raise RuntimeError(f'{missing} are not implemented')
    return service



from orjson import loads, dumps
from pymaid.error import ErrorManager
from pymaid.rpc.error import RPCError
from pymaid.rpc.context import InboundContext, OutboundContext

from .pymaid_pb2 import Context as Meta, ErrorMessage, Void


class PBContext:

    async def handle_error(self, error: Exception):
        await self.conn.send_message(
            Meta(
                transmission_id=self.transmission_id,
                packet_flags=(
                    self.method.options.get('flags', 0) | Meta.PacketFlag.END
                ),
                is_failed=True,
            ),
            ErrorMessage(
                code=error.code,
                message=error.message,
                data=dumps(error.data) if error.data else '',
            ),
        )
        self.sent_end_message = True

    async def shutdown(self):
        await self.conn.send_message(
            Meta(
                transmission_id=self.transmission_id,
                packet_flags=(
                    self.method.options.get('flags', 0) | Meta.PacketFlag.END
                ),
            ),
            Void(),
        )


class PBInboundContext(PBContext, InboundContext):

    def feed_message(self, meta, payload):
        '''Received request from transport layer'''
        assert meta.packet_type == Meta.REQUEST, f'invalid {meta.packet_type=}'
        if self.request_fed_count > 0 and not self.method.client_streaming:
            raise RPCError.MultipleRequestForUnaryMethod(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )
        if payload:
            self.request_queue.append(
                self.method.request_class.FromString(payload)
            )
        if meta.packet_flags & Meta.PacketFlag.END:
            self.request_queue.append(None)
        self.request_fed_count += 1
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(True)

    async def send_message(
        self, response=None, *, end: bool = False, **kwargs
    ):
        '''Send response to transport layer'''
        if self.method.options.get('void_response'):
            # do not send_message when response is void
            return
        if self.response_sent_count > 0 and not self.method.server_streaming:
            raise RPCError.RPCResponseSent(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )
        flags = self.method.options.get('flags', 0)
        if end or not self.method.server_streaming:
            flags |= Meta.PacketFlag.END
            self.sent_end_message = True
        await self.conn.send_message(
            Meta(
                transmission_id=self.transmission_id,
                packet_type=Meta.RESPONSE,
                packet_flags=flags
            ),
            response or self.method.response_class(**kwargs),
        )


class PBOutboundContext(PBContext, OutboundContext):

    def feed_message(self, meta, payload):
        '''Received response from transport layer '''
        assert not meta.packet_type or meta.packet_type == Meta.RESPONSE, \
            f'invalid {meta.packet_type=}'
        if self.response_fed_count > 0 and not self.method.server_streaming:
            raise RPCError.MultipleResponseForUnaryMethod(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )

        if meta.is_failed or meta.is_cancelled:
            assert payload, 'should return error message'
            err = ErrorMessage.FromString(payload)
            ex = ErrorManager.assemble(
                err.code, err.message, err.data and loads(err.data) or {}
            )
            self.response_queue.append(ex)
        elif payload:
            self.response_queue.append(
                self.method.response_class.FromString(payload)
            )
        if meta.packet_flags & Meta.PacketFlag.END:
            self.response_queue.append(None)
        if self.waiter and not self.waiter.done():
            self.waiter.set_result(True)

    async def send_message(self, request=None, *, end: bool = False, **kwargs):
        '''Send request to transport layer'''
        if self.request_sent_count > 0 and not self.method.client_streaming:
            raise RPCError.RPCRequestSent(
                data={
                    'service_method': self.method.full_name,
                    'transmission_id': self.transmission_id,
                }
            )

        flags = self.method.options.get('flags', 0)
        if end or not self.method.client_streaming:
            flags |= Meta.PacketFlag.END
            self.sent_end_message = True
        await self.conn.send_message(
            Meta(
                transmission_id=self.transmission_id,
                service_method=self.method.full_name,
                packet_type=Meta.REQUEST,
                packet_flags=flags,
            ),
            request or self.method.request_class(**kwargs),
        )



# inspired by https://github.com/aio-libs/async-timeout

import asyncio
import enum
from types import TracebackType
from typing import Optional, Type

from pymaid.core import current_task, get_running_loop


__all__ = ('timeout', 'timeout_at')


def timeout(delay: Optional[float]) -> 'Timeout':
    '''timeout context manager.

    Useful in cases when you want to apply timeout logic around block
    of code or in cases when asyncio.wait_for is not suitable. For example:

    >>> async with timeout(0.001):
    ...     async with aiohttp.get('https://github.com') as r:
    ...         await r.text()


    delay - value in seconds or None to disable timeout logic
    '''
    if delay is not None:
        deadline = get_running_loop().time() + delay  # type: Optional[float]
    else:
        deadline = None
    return Timeout(deadline)


def timeout_at(deadline: Optional[float]) -> 'Timeout':
    '''Schedule the timeout at absolute time.

    deadline arguments points on the time in the same clock system
    as loop.time().

    Please note: it is not POSIX time but a time with
    undefined starting base, e.g. the time of the system power on.

    >>> async with timeout_at(loop.time() + 10):
    ...     async with aiohttp.get('https://github.com') as r:
    ...         await r.text()


    '''
    return Timeout(deadline)


class State(enum.Enum):
    INIT = 'INIT'
    ENTER = 'ENTER'
    TIMEOUT = 'TIMEOUT'
    EXIT = 'EXIT'


class Timeout:
    # Internal class, please don't instantiate it directly
    # Use timeout() and timeout_at() public factories instead.
    #
    # Implementation note: `async with timeout()` is preferred
    # over `with timeout()`.
    # While technically the Timeout class implementation
    # doesn't need to be async at all,
    # the `async with` statement explicitly points that
    # the context manager should be used from async function context.
    #
    # This design allows to avoid many silly misusages.
    #
    # TimeoutError is raised immadiatelly when scheduled
    # if the deadline is passed.
    # The purpose is to time out as sson as possible
    # without waiting for the next await expression.

    __slots__ = ('deadline', 'loop', 'state', 'task', 'timeout_handler')

    def __init__(self, deadline: Optional[float]):
        self.loop = get_running_loop()
        self.task = current_task()
        self.state = State.INIT

        self.timeout_handler = None  # type: Optional[asyncio.Handle]
        if deadline is None:
            self.deadline = None  # type: Optional[float]
        else:
            self.shift_to(deadline)

    async def __aenter__(self) -> 'Timeout':
        if self.state != State.INIT:
            raise RuntimeError(f'invalid state {self.state.value}')
        self.state = State.ENTER
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> Optional[bool]:
        if exc_type is asyncio.CancelledError and self.state == State.TIMEOUT:
            self.timeout_handler = None
            raise asyncio.TimeoutError
        # timeout is not expired
        self.state = State.EXIT
        self._abort()
        return None

    @property
    def expired(self) -> bool:
        '''Is timeout expired during execution?'''
        return self.state == State.TIMEOUT

    def abort(self) -> None:
        '''Abort scheduled timeout if any.'''
        # cancel is maybe better name but
        # task.cancel() raises CancelledError in asyncio world.
        if self.state not in (State.INIT, State.ENTER):
            raise RuntimeError(f'invalid state {self.state.value}')
        self._abort()

    def _abort(self) -> None:
        if self.timeout_handler is not None:
            self.timeout_handler.cancel()
            self.timeout_handler = None

    def shift_by(self, delay: float) -> None:
        '''Advance timeout on delay seconds.

        The delay can be negative.
        '''
        self.shift_to(self.loop.time() + delay)

    def shift_to(self, deadline: float) -> None:
        '''Set timeout deadline to the absolute time.

        If new deadline is in the past, the timeout is raised immediately.
        '''
        if self.state == State.EXIT:
            raise RuntimeError(
                'cannot reschedule after exit from context manager'
            )
        if self.state == State.TIMEOUT:
            raise RuntimeError('cannot reschedule expired timeout')
        if self.timeout_handler is not None:
            self.timeout_handler.cancel()
        self.deadline = deadline
        now = self.loop.time()
        if deadline <= now:
            self.timeout_handler = None
            if self.state == State.INIT:
                raise asyncio.TimeoutError
            else:
                # state is ENTER
                raise asyncio.CancelledError
        self.timeout_handler = self.loop.call_at(
            deadline, self.on_timeout, self.task
        )

    def on_timeout(self, task: 'asyncio.Task[None]') -> None:
        task.cancel()
        self.state = State.TIMEOUT



import struct

from base64 import b64encode
from functools import partial
from hashlib import sha1
from os import urandom
from socket import error as socket_error

from pymaid.conf import settings
from pymaid.core import create_task, BaseTransport, Event
from pymaid.net.stream import Stream
from pymaid.types import DataType
from pymaid.utils.logger import logger_wrapper

from .exceptions import ProtocolError, WebSocketError
from .protocol import unmask_payload, WSProtocol, Header
from .utf8validator import Utf8Validator


@logger_wrapper
class WebSocket(Stream):

    def __init__(self, *, server=None, initiative=False, resource='/'):
        super().__init__(server=server, initiative=initiative)
        self.utf8validator = Utf8Validator()
        self.timeout = settings.namespaces['pymaid']['PM_WEBSOCKET_TIMEOUT']
        self.resource = resource.encode('utf-8')
        self.protocol = WSProtocol()

    def _done_handshake(self, task):
        self.conn_made_event.set()
        if self.initiative:
            self.conn = self.server.connection_made(self)

    def _read_headers(self):
        headers = {}
        raw_readline = super(WebSocket, self).readline
        while 1:
            line = raw_readline(1024, self.timeout).strip()
            if not line:
                break
            key, value = line.split(b':', 1)
            headers[key] = value.strip()
        return headers

    def _do_handshake(self):
        self.logger.debug(f'{self.conn_id} doing handshake')
        key = b64encode(urandom(16)).strip()
        self.transport.write(self.protocol.handshake_frame(self.resource, key))
        line = super(WebSocket, self).readline(1024, self.timeout)
        if not line:
            self.close(reset=True)
            return False
        status = line.split(b' ', 2)
        if len(status) != 3:
            self.close('invalid response line')
            return False
        status = int(status[1])
        if status != 101:
            self.close(f'handshake failed with status: {status}')
            return False

        headers = self._read_headers()

        if (b'websocket' != headers.get(b'Upgrade')
                or b'Upgrade' != headers.get(b'Connection')):
            self.close('invalid websocket handshake header')
            return False

        accept = headers.get(b'Sec-WebSocket-Accept', '').lower()
        if isinstance(accept, str):
            accept = accept.encode('utf-8')
        value = key + b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
        hashed = b64encode(sha1(value).digest()).strip().lower()
        if hashed != accept:
            self.close('invalid accept value')
            return False
        return True

    def _upgrade_connection(self):
        line = super(WebSocket, self).readline(1024, self.timeout)
        if not line:
            self.close(reset=True)
            return False

        datas = line.split()
        if len(datas) != 3:
            self.close(f'invalid request line: {line}')
            return False
        method, path, version = datas
        if method != b'GET' or version != b'HTTP/1.1':
            self.close(f'websocket requried GET HTTP/1.1, got `{line}`')
            return False

        headers = self._read_headers()

        sec_key = headers.get(b'Sec-WebSocket-Key', '')
        if not sec_key:
            self.close('no Sec-WebSocket-Key')
            return False

        version = headers.get(b'Sec-WebSocket-Version')
        if version not in self.SUPPORTED_VERSIONS:
            self.close(f'not supported version: {version}')
            return False

        resp_key = b64encode(sha1(sec_key + self.GUID).digest())
        self._send_queue.append(self.HANDSHAKE_RESP % resp_key)
        self._send()
        return True

    def _write(self, opcode: int, payload: DataType):
        self._send()

    def connection_made(self, transport: BaseTransport):
        self.__class__.CONN_ID = self.__class__.CONN_ID + 1
        self.conn_id = f'{self.__class__.__name__}-{self.__class__.CONN_ID}'
        self.bind_transport(transport)
        if self.initiative:
            self.conn_made_event = Event()
            self.handshake_handler = create_task(self._do_handshake())
            self.handshake_handler.add_done_callback(self.done_handshake)
        elif self.server:
            self.conn = self.server.connection_made(self)
        self.logger.info(f'[{self}] made')

    def oninit(self):
        ''' Called by handler once handler start on a greenlet.'''
        if self.initiative:
            self.connecting_event.wait()
            return self.is_connected
        else:
            return self._upgrade_connection()

    def handle_close(self, header, payload):
        '''Called when a close frame has been decoded from the stream.

        :param header: The decoded `Header`.
        :param payload: The bytestring payload associated with the close frame.
        '''
        if not payload:
            self.close(1000, None)
            return

        if len(payload) < 2:
            raise ProtocolError(f'Invalid close frame: {header} {payload}')

        code = struct.unpack('!H', payload[:2])[0]
        payload = payload[2:]

        if payload:
            val = self.utf8validator.validate(payload)
            if not val[0]:
                raise UnicodeError

        if (code < 1000 or 1004 <= code <= 1006 or 1012 <= code <= 1016
                or code == 1100 or 2000 <= code <= 2999):
            raise ProtocolError(f'Invalid close code {code}')
        self.close(code, payload)

    def handle_ping(self, header, payload):
        self.send_frame(payload, self.OPCODE_PONG)

    def handle_pong(self, header, payload):
        pass

    def read_frame(self):
        '''Block until a full frame has been read from the socket.

        This is an internal method as calling this will not cleanup correctly
        if an exception is called. Use `receive` instead.

        :return: The header and payload as a tuple.
        '''
        header = Header.decode_header(
            partial(super(WebSocket, self).read, timeout=self.timeout)
        )
        if not header:
            return header, b''

        if header.flags:
            raise ProtocolError(f'invalid flags: {header.flags}')

        if not header.length:
            return header, b''

        payload = super(WebSocket, self).read(header.length, self.timeout)
        if len(payload) != header.length:
            raise WebSocketError('Unexpected EOF reading frame payload')

        if header.mask:
            payload = unmask_payload(payload, header.mask)
        return header, payload

    def read_message(self):
        '''Return the next text or binary message from the socket.

        This is an internal method as calling this will not cleanup correctly
        if an exception is called. Use `receive` instead.
        '''
        opcode = None
        message = b''

        while 1:
            header, payload = self.read_frame()
            if not header:
                break

            f_opcode = header.opcode
            if f_opcode in (self.OPCODE_TEXT, self.OPCODE_BINARY):
                # a new frame
                if opcode:
                    raise ProtocolError(
                        f'The opcode in non-fin frame is expected to be zero,'
                        f'got {f_opcode}'
                    )

                # Start reading a new message, reset the validator
                self.utf8validator.reset()
                opcode = f_opcode
            elif f_opcode == self.OPCODE_CONTINUATION:
                if not opcode:
                    raise ProtocolError('Unexpected frame with opcode=0')
            elif f_opcode == self.OPCODE_PING:
                self.handle_ping(header, payload)
                continue
            elif f_opcode == self.OPCODE_PONG:
                self.handle_pong(header, payload)
                continue
            elif f_opcode == self.OPCODE_CLOSE:
                self.handle_close(header, payload)
                return
            else:
                raise ProtocolError(f'Unexpected opcode={f_opcode}')

            message += payload
            if header.fin:
                break

        if opcode == self.OPCODE_TEXT:
            stat = self.utf8validator.validate(message)
            if not stat[0]:
                raise UnicodeError(
                    f'Encountered invalid UTF-8 while processing '
                    f'text message at payload octet index {stat[3]}'
                )
            return message
        else:
            return message

    def receive(self):
        '''Read and return a message from the stream.

        If `None` is returned, then the socket is considered closed/errored.
        '''

        try:
            return self.read_message()
        except UnicodeError:
            self.close(1007)
        except ProtocolError:
            self.close(1002)
        except socket_error:
            self.close()



import logging

from functools import wraps
from logging import LogRecord
from sys import _getframe as getframe
from time import time
from types import MethodType

from google.protobuf.service import Service

logging_levels = logging._levelToName
logging_names = logging._nameToLevel


def update_record(record, level, msg, *args):
    record.levelno = level
    record.levelname = logging_levels[level]
    record.msg = msg
    record.args = args
    ct = time()
    record.created = ct
    record.msecs = (ct - int(ct)) * 1000


def trace_service(
    level=logging.INFO,
    debug_info_func=lambda ctx: f'[conn|{ctx.conn.conn_id}]'
):
    def wrapper(cls):
        assert level in logging_levels, (level, logging_levels)
        for method in cls.DESCRIPTOR.methods:
            name = method.name
            setattr(
                cls,
                name,
                trace_method(level, debug_info_func)(getattr(cls, name))
            )
        return cls
    if isinstance(level, str):
        level = logging_names[level]
        return wrapper
    elif isinstance(level, int):
        return wrapper
    else:
        assert issubclass(level, Service), level
        cls, level = level, logging.INFO
        return wrapper(cls)


def trace_method(
    level=logging.INFO,
    debug_info_func=lambda ctx: '[conn|%d]' % ctx.conn.conn_id
):
    def wrapper(func):
        assert level in logging_levels, (level, logging_levels)
        co = func.__code__
        method = co.co_name
        if isinstance(func, MethodType):
            method = f'{func.im_class.DESCRIPTOR.name}.{method}'

        # name, level, fn, lno, msg, args, exc_info, func
        record = logging.LogRecord(
            '', level, co.co_filename, co.co_firstlineno, '', (), None, method
        )

        @wraps(func)
        async def _(self, request, context):
            assert isinstance(self, Service)

            start_time = time()
            debug_info = debug_info_func(context)
            record.name = self.logger.name
            req = repr(str(request))

            try:
                await func(self, request, context)
            except BaseException as ex:
                if isinstance(ex, Warning):
                    update_record(
                        record,
                        logging.WARN,
                        '%s [rpc|%s] [req|%.1024r] [warning|%s] [time|%.6f]',
                        debug_info,
                        method,
                        req,
                        ex,
                        time() - start_time,
                    )
                else:
                    update_record(
                        record,
                        logging.ERROR,
                        '%s [rpc|%s] [req|%.1024r] [exception|%s] [time|%.6f]',
                        debug_info,
                        method,
                        req,
                        ex,
                        time() - start_time,
                    )
                self.logger.handle(record)
                raise
            update_record(
                record,
                level,
                '%s [rpc|%s] [req|%.1024r] [time|%.6f]',
                debug_info,
                method,
                req,
                time() - start_time,
            )
            self.logger.handle(record)
        return _
    if isinstance(level, str):
        level = logging_names[level]
        return wrapper
    elif isinstance(level, int):
        return wrapper
    else:
        assert callable(level), level
        func, level = level, logging.INFO
        return wrapper(func)


def trace_stub(level=logging.DEBUG):
    def wrapper(method):
        from pymaid.conf import settings
        if not settings.get('DEBUG', False, ns='pymaid'):
            return method
        assert level in logging_levels, (level, logging_levels)

        @wraps(method)
        def _(
            self,
            request=None,
            *,
            transport=None,
            broadcaster=None,
            protocol=None,
            **kwargs
        ):
            frame = getframe(1)
            method.logger.handle(LogRecord(
                method.logger.name,
                level,
                frame.f_code.co_filename,
                frame.f_lineno,
                '[stub|%s][request|%r][kwargs|%r]',
                (method.name, str(request), str(kwargs)),
                None,
                method.name,
            ))
            return method(
                self, request, transport, broadcaster, protocol, **kwargs
            )
        return _
    if isinstance(level, str):
        level = logging_names[level]
        return wrapper
    elif isinstance(level, int):
        return wrapper
    else:
        assert callable(level), level
        method, level = level, logging.DEBUG
        return wrapper(method)



from pymaid.rpc.error import RPCError


PBError = RPCError.create_manager('PBError')
PBError.add_error('RPCNotFound', 'rpc not found')
PBError.add_error(
    'HeartbeatTimeout', '[host|{}][peer|{}] peer heartbeat timeout'
)
PBError.add_error('PacketTooLarge', 'packet_length out of limitation')
PBError.add_error('EOF', 'socket received eof')
PBError.add_error('Timeout', 'socket read/readline timeout {}')
PBError.add_error('InvalidTransmissionID', 'transmission_id value is invalid')
PBError.add_error('InvalidPacketType', 'cannot handle unknown packet')



'''Pymaid is based on asyncio now.

And for better performance, it use uvloop as the event loop
'''
import asyncio
import socket

from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial

__all__ = (
    'run',
    'sleep',
    'get_event_loop',
    'get_event_loop_policy',
    'get_running_loop',

    'create_task',
    'current_task',
    'all_tasks',
    'wait_for',
    'wait',
    'gather',
    'Task',
    'TimeoutError',
    'CancelledError',
    'Future',
    'Event',
    'Semaphore',

    'Queue',
    'LifoQueue',
    'PriorityQueue',
    'QueueFull',
    'QueueEmpty',

    'create_stream',
    'create_datagram',
    'create_stream_server',
    'create_datagram_server',
    'create_unix_stream',
    'create_unix_stream_server',

    'iscoroutine',
    'iscoroutinefunction',
    'ensure_future',
    'wrap_future',
    'run_in_threadpool',
    'run_in_processpool',
)


#
# event loop
#

run = asyncio.run
sleep = asyncio.sleep
get_event_loop = asyncio.get_event_loop
get_event_loop_policy = asyncio.get_event_loop_policy
get_running_loop = asyncio.get_running_loop


#
# tasks
#

create_task = asyncio.create_task
current_task = asyncio.current_task
all_tasks = asyncio.all_tasks

wait_for = asyncio.wait_for
wait = asyncio.wait
gather = asyncio.gather

Task = asyncio.Task

TimeoutError = asyncio.TimeoutError
CancelledError = asyncio.CancelledError

Future = asyncio.Future
Event = asyncio.Event
Semaphore = asyncio.Semaphore


#
# queue
#

Queue = asyncio.Queue
LifoQueue = asyncio.LifoQueue
PriorityQueue = asyncio.PriorityQueue
QueueFull = asyncio.QueueFull
QueueEmpty = asyncio.QueueEmpty


#
# network
#

BaseTransport = asyncio.BaseTransport
BaseProtocol = asyncio.BaseProtocol


async def create_stream(transport_class, host, port, **kwargs):
    return (await get_running_loop().create_connection(
        partial(transport_class, initiative=True), host, port, **kwargs
    ))[1]


async def create_datagram(transport_class, addr, **kwargs):
    return (await get_running_loop().create_datagram_endpoint(
        partial(transport_class, initiative=True), remote_addr=addr, **kwargs
    ))[1]


async def create_stream_server(transport_class, host, port, **kwargs):
    return await get_running_loop().create_server(
        transport_class, host, port, **kwargs
    )


async def create_datagram_server(transport_class, addr, **kwargs):
    return await get_running_loop().create_datagram_endpoint(
        transport_class, local_addr=addr, **kwargs
    )


async def create_unix_stream(transport_class, path, **kwargs):
    return (await get_running_loop().create_unix_connection(
        partial(transport_class, initiative=True), path, **kwargs
    ))[1]


async def create_unix_datagram(transport_class, path, **kwargs):
    return (await get_running_loop().create_datagram_endpoint(
        partial(transport_class, initiative=True),
        remote_addr=path,
        family=socket.AF_UNIX,
        **kwargs
    ))[1]


async def create_unix_stream_server(transport_class, path, **kwargs):
    return await get_running_loop().create_unix_server(
        transport_class, path, **kwargs
    )


async def getaddrinfo(host, port, family=0, type=0, proto=0, flags=0):
    return await run_in_threadpool(
        socket.getaddrinfo, args=(host, port, family, type, proto, flags),
    )


# unix domain socket not support datagram

# async def create_unix_datagram_server(protocol, path, **kwargs):
#     loop = get_running_loop()
#     return await loop.create_datagram_endpoint(
#         protocol, local_addr=path, family=socket.AF_UNIX, **kwargs
#     )


#
# utils
#

iscoroutine = asyncio.iscoroutine
iscoroutinefunction = asyncio.iscoroutinefunction
ensure_future = asyncio.ensure_future
wrap_future = asyncio.futures.wrap_future


# executor

default_thread_executor = ThreadPoolExecutor()
default_process_executor = ProcessPoolExecutor()


def run_in_threadpool(
    func, *, args=None, kwargs=None, executor=default_thread_executor,
):
    return wrap_future(executor.submit(func, *(args or ()), **(kwargs or {})))


def run_in_processpool(
    func, *, args=None, kwargs=None, executor=default_process_executor,
):
    return wrap_future(executor.submit(func, *(args or ()), **(kwargs or {})))


del asyncio



import abc
from queue import deque
from typing import Optional, Union

from pymaid.core import create_task, Event
from pymaid.error import BaseEx
from pymaid.ext.pools.worker import AioPool
from pymaid.utils.logger import logger_wrapper

from .context import C, InboundContext, OutboundContext
from .error import RPCError
from .method import Method, MethodStub
from .router import Router


@logger_wrapper
class Handler(abc.ABC):
    '''Handle the *received* request/response.'''

    MAX_TRANSMISSION_ID = 2 ** 32 - 1
    INBOUND_CONTEXT_CLASS = InboundContext
    OUTBOUND_CONTEXT_CLASS = OutboundContext

    def __init__(
        self,
        router: Router,
        timeout: Optional[float] = None,
    ):
        self.conn = None
        self.task = None
        self.router = router
        self.timeout = timeout

        # self.inbound_transmission_id = 0
        self.outbound_transmission_id = 0
        self.contexts = {}

        self.pending_tasks = deque()
        self.new_task_received = Event()
        self.closed_event = Event()
        self.is_closing = False
        self.is_closed = False

    def start(self, conn):
        # NOTE: cyclic
        self.conn = conn

        # for initiative side, the id will be EVEN
        # for passive side, the id will be ODD
        if conn.initiative:
            self.outbound_transmission_id = 1
        else:
            self.outbound_transmission_id = 2
        # cyclic
        self.task = create_task(self.run())

    def get_context(self, transmission_id: int) -> 'C':
        return self.contexts.get(transmission_id)

    def release_context(self, transmission_id: int):
        if transmission_id not in self.contexts:
            # already released ?
            return
        context = self.contexts.pop(transmission_id)
        del context._manager

    def next_transmission_id(self) -> int:
        '''Return the next available transmission id for the context created
        by this endpoint.

        :raises: :class:`NoAvailableTransmissionID
            <pymaid.rpc.error.RPCError.NoAvailableTransmissionID>`
        :returns: the next id can be initiate a Context
        :rtype: ``int``
        '''
        transmission_id = self.outbound_transmission_id
        if transmission_id > self.MAX_TRANSMISSION_ID:
            raise RPCError.NoAvailableTransmissionID(
                data={'id': transmission_id, 'max': self.MAX_TRANSMISSION_ID}
            )
        self.outbound_transmission_id += 2
        return transmission_id

    def new_inbound_context(
        self,
        transmission_id: int,
        *,
        method: Method,
        timeout: Optional[float] = None,
    ) -> 'C':
        # it is hard to insist the inbound_transmission_id order
        # e.g.
        # client make 3 async calls: 1, 3, 5
        # all 3 run parallelly, then the order is unspecified

        # if transmission_id < self.inbound_transmission_id:
        #     raise RPCError.InvalidTransmissionID(
        #         data={
        #             'tid': transmission_id,
        #             'reason': 'reused transmission id',
        #         }
        #     )

        transmission_id = transmission_id
        assert transmission_id not in self.contexts, 'reused transmission id'
        if not self.conn.initiative and transmission_id % 2 != 1:
            raise RPCError.InvalidTransmissionID(
                data={
                    'tid': transmission_id,
                    'reason': 'invalid inbound transmission id value',
                }
            )
        # self.inbound_transmission_id = transmission_id

        # warning: cyclic referrence
        context = self.INBOUND_CONTEXT_CLASS(
            conn=self.conn,
            transmission_id=transmission_id,
            method=method,
            timeout=timeout,
        )
        context._manager = self
        self.contexts[transmission_id] = context
        return context

    def new_outbound_context(
        self,
        *,
        method: MethodStub,
        timeout: Optional[float] = None,
    ) -> 'C':
        transmission_id = self.next_transmission_id()
        # warning: cyclic referrence
        context = self.OUTBOUND_CONTEXT_CLASS(
            conn=self.conn,
            transmission_id=transmission_id,
            method=method,
            timeout=timeout,
        )
        context._manager = self
        self.contexts[transmission_id] = context
        return context

    @abc.abstractmethod
    async def run(self):
        raise NotImplementedError('run')

    def shutdown(self, reason: Union[None, str, Exception] = None):
        if self.is_closing:
            return
        self.is_closing = True
        self.pending_tasks.append(None)
        self.new_task_received.set()

    async def join(self, reason: Optional[Union[str, Exception]] = None):
        await self.closed_event.wait()

    def close(self, reason: Optional[Union[str, Exception]] = None):
        if self.is_closed:
            return
        self.is_closed = True
        self.pending_tasks.clear()
        self.closed_event.set()
        self.conn.close(reason)
        if not self.task.done():
            self.task.cancel()
        del self.conn
        del self.task

    @abc.abstractmethod
    def feed_messages(self, messages):
        # self.logger.debug(f'{self} feed {len(messages)=}')
        raise NotImplementedError('feed_messages')

    async def handle_error(self, error):
        '''Default error handler.

        Just write error logs.
        '''
        self.logger.error(f'{self!r} caught an unhandled error, {error=!r}')

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'pending={len(self.pending_tasks)}>'
        )


@logger_wrapper
class SerialHandler(Handler):
    '''Handle the *received* requests *one by one*.

    Even for streaming method, it will still in serial order.

    This is usually used for stateful situation, like gaming, the previous
    request mostly will impact the later one.
    '''

    async def run(self):
        new_task_received = self.new_task_received
        pending_tasks = self.pending_tasks

        while 1:
            await new_task_received.wait()
            while pending_tasks:
                task = pending_tasks.popleft()
                if not task:
                    self.close()
                    return
                try:
                    await task
                except BaseEx as exc:
                    assert False, f'{exc} should be handled within logic'
                except Exception as exc:
                    self.close(exc)
                    return
            new_task_received.clear()


@logger_wrapper
class ParallelHandler(Handler):
    '''Handle the *received* requests parallelly.

    It holds a worker pool to do the actual work, with a limited concurrency.
    '''

    def __init__(
        self,
        router: Router,
        timeout: Optional[float] = None,
        concurrency: int = 5,
    ):
        super().__init__(router, timeout)
        self.worker = AioPool(concurrency)

    async def join(self, reason: Optional[Union[str, Exception]] = None):
        await self.worker.shutdown(wait=True)
        await super().join(reason)

    async def run(self):
        new_task_received = self.new_task_received
        pending_tasks = self.pending_tasks
        # run_task = self.worker.submit
        run_task = self.worker.spawn

        while 1:
            await new_task_received.wait()
            while pending_tasks:
                task = pending_tasks.popleft()
                if not task:
                    self.close()
                    return
                try:
                    await run_task(task)
                except BaseEx as exc:
                    assert False, f'{exc} should be handled within logic'
                except Exception as exc:
                    self.close(exc)
                    return
            new_task_received.clear()



from .base import ErrorManager


RpcError = ErrorManager.create_manager('RpcError')
RpcError.add_error('RPCNotFound', 'rpc not found')
RpcError.add_error(
    'HeartbeatTimeout', '[host|{}][peer|{}] peer heartbeat timeout'
)
RpcError.add_error('PacketTooLarge', '[packet_length|{}] out of limitation')
RpcError.add_error('EOF', 'socket received eof')
RpcError.add_error('Timeout', 'socket read/readline timeout {}')



import os
import socket

import pytest

from pymaid.core import sleep
from pymaid.net.raw import HAS_IPv6_FAMILY

from tests.common.models import _TestStream


@pytest.mark.skipif(
    os.name == 'posix', reason='linux system does not support INET4 socketpair'
)
@pytest.mark.asyncio
async def test_stream_ipv4():
    sock1, sock2 = socket.socketpair(socket.AF_INET.value)
    s1, s2 = _TestStream(sock1), _TestStream(sock2)
    await s1.write(b'from pymaid')
    await sleep(0)
    assert s2.received_data == b'from pymaid'
    s1.close()
    s2.close()


@pytest.mark.skipif(
    os.name == 'posix', reason='linux system does not support INET6 socketpair'
)
@pytest.mark.skipif(not HAS_IPv6_FAMILY, reason='does not support ipv6')
@pytest.mark.asyncio
async def test_stream_ipv6():
    sock1, sock2 = socket.socketpair(socket.AF_INET6.value)
    s1, s2 = _TestStream(sock1), _TestStream(sock2)
    await s1.write(b'from pymaid')
    await sleep(0)
    assert s2.received_data == b'from pymaid'
    s1.close()
    s2.close()


@pytest.mark.skipif(
    not getattr(socket, 'AF_UNIX', None),
    reason='does not support unix domain sock'
)
@pytest.mark.asyncio
async def test_stream_unix():
    sock1, sock2 = socket.socketpair(socket.AF_UNIX.value)
    s1, s2 = _TestStream(sock1), _TestStream(sock2)
    await s1.write(b'from pymaid')
    await sleep(0)
    assert s2.received_data == b'from pymaid'
    s1.close()
    s2.close()



import abc
import os
import socket
import ssl as _ssl
import sys

from typing import Callable, List, Optional, Tuple, TypeVar, Union

from pymaid.conf import settings
from pymaid.core import get_running_loop, CancelledError, Event
from pymaid.ext.middleware import MiddlewareManager

from .base import logger, ChannelState
from .raw import sock_connect, sock_listen
from .stream import Stream, StreamType
from .transport import Transport, TransportType


class Channel(abc.ABC):

    STATE = ChannelState
    logger = logger

    def __init__(
        self,
        *,
        name: str = 'Channel',
        address: Union[Tuple[str, int], str] = '',
        transport_class: TransportType = Transport,
        ssl_context: _ssl.SSLContext,
        ssl_handshake_timeout: Optional[float] = None,
        middleware_manager: Optional[MiddlewareManager] = None,
    ):
        '''Channel manages the sockets.'''
        self.name = name
        self.address = address
        self.transport_class = transport_class
        self.ssl_context = ssl_context
        self.ssl_handshake_timeout = ssl_handshake_timeout
        self.transports = {}
        self.middleware_manager = middleware_manager or MiddlewareManager()

        self.listeners = []
        self.state = self.STATE.CREATED
        self.closed_event = Event()
        self._loop = get_running_loop()
        self._serving_forever_fut = None

    @property
    def is_full(self):
        return len(self.transports) >= settings.pymaid.MAX_CONNECTIONS

    async def listen(
        self,
        address: Union[Tuple[str, int], str],
        *,
        family: socket.AddressFamily = socket.AF_UNSPEC,
        flags: socket.AddressInfo = socket.AI_PASSIVE,
        backlog: int = 128,
        reuse_address: bool = os.name == 'posix' and sys.platform != 'cygwin',
        reuse_port: bool = False,
    ):
        listeners = await sock_listen(
            address,
            family=family,
            flags=flags,
            backlog=backlog,
            reuse_address=reuse_address,
            reuse_port=reuse_port,
        )
        for sock in listeners:
            self.listeners.append(sock)

    def read_from_listener(self, sock: socket.socket, backlog: int = 128):
        raise NotImplementedError

    async def wait_closed(self):
        await self.closed_event.wait()

    async def serve_forever(self):
        if self._serving_forever_fut is not None:
            raise RuntimeError(
                f'channel {self!r} is already being awaited on serve_forever()'
            )
        if not self.listeners:
            raise RuntimeError(f'channel {self!r} has no listeners')

        self._serving_forever_fut = self._loop.create_future()

        try:
            await self._serving_forever_fut
        except CancelledError:
            try:
                self.close()
            finally:
                raise
        finally:
            self._serving_forever_fut = None

    def start(self):
        if self.state >= self.STATE.CLOSING:
            raise RuntimeError(f'{self!r} is closing, cannot start again')
        self.logger.info(f'{self!r} start')
        self.state = self.STATE.STARTED
        loop = self._loop
        for sock in self.listeners:
            loop.add_reader(sock.fileno(), self.read_from_listener, sock)

    def pause(self, reason: str = ''):
        self.logger.info(f'{self!r} pause with reason: {reason}')
        self.state = self.STATE.PAUSED
        loop = self._loop
        for sock in self.listeners:
            loop.remove_reader(sock.fileno())

    def shutdown(self, reason: str = 'shutdown'):
        if self.state >= self.STATE.CLOSING:
            return
        if self.state < self.STATE.PAUSED:
            self.pause(reason)
        self.state = self.STATE.CLOSING
        self.logger.info(f'{self!r} shutdown with reason: {reason}')

    def close(
        self, reason: Union[None, str, Exception] = 'called close',
    ):
        if self.state >= self.STATE.CLOSING:
            return
        if self.state == self.STATE.STARTED:
            self.pause(reason)
        if self.state == self.STATE.PAUSED:
            self.shutdown(reason)
        for sock in self.listeners:
            sock.close()
        del self.listeners[:]
        if not self.transports:
            self._finnal_close(reason)

    def _finnal_close(self, reason=None):
        self.closed_event.set()
        self.state = self.STATE.CLOSED
        self.logger.info(f'{self!r} finally closed, {reason=}')

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()
        self.close(exc_val)
        await self.wait_closed()
        if exc_val:
            raise exc_val

    def __repr__(self):
        return (
            '<'
            f'{self.name} '
            f'state={self.state.name} '
            f'listeners={len(self.listeners)} '
            f'transports={len(self.transports)} '
            f'middlewares={len(self.middleware_manager.middlewares)}'
            '>'
        )


class StreamChannel(Channel):

    def __init__(
        self,
        *,
        name: str = 'Channel',
        address: Union[Tuple[str, int], str] = '',
        transport_class: StreamType = Stream,
        ssl_context: _ssl.SSLContext,
        ssl_handshake_timeout: Optional[float] = None,
        middleware_manager: Optional[MiddlewareManager] = None,
    ):
        super().__init__(
            name=name,
            address=address,
            transport_class=transport_class,
            ssl_context=ssl_context,
            ssl_handshake_timeout=ssl_handshake_timeout,
            middleware_manager=middleware_manager,
        )

    def read_from_listener(self, sock: socket.socket, backlog: int = 128):
        connection_made = self.connection_made
        for _ in range(backlog):
            if self.is_full:
                self.pause('{self!r} stop accept since is full')
                break
            try:
                conn, addr = sock.accept()
            except (BlockingIOError, InterruptedError, ConnectionAbortedError):
                return
            conn.setblocking(False)
            connection_made(conn)

    async def acquire(
        self,
        *,
        on_open: Optional[List[Callable]] = None,
        on_close: Optional[List[Callable]] = None,
    ) -> Stream:
        sock = await sock_connect(self.address)
        conn = self._make_connection(sock, True, on_open, on_close)
        self.logger.info(
            f'{self!r} acquire: <{self.transport_class.__name__} {conn.id}>'
        )
        return conn

    def connection_made(self, sock: socket.socket) -> Stream:
        conn = self._make_connection(
            sock, False, on_close=[self.connection_lost],
        )
        self.logger.info(
            f'{self!r} connection_made: '
            f'<{self.transport_class.__name__} {conn.id}>'
        )
        self.transports[conn.id] = conn
        return conn

    def connection_lost(self, conn: Stream, exc=None):
        self.logger.info(
            f'{self!r} connection_lost: '
            f'<{self.transport_class.__name__} {conn.id}> {exc=}'
        )
        assert conn.id in self.transports, conn.id
        del self.transports[conn.id]
        if self.state == self.STATE.PAUSED and not self.is_full:
            self.start()
        if not self.transports and self.state >= self.STATE.CLOSING:
            self._finnal_close(exc)

    def shutdown(self, reason: str = 'shutdown'):
        super().shutdown(reason)
        for conn in self.transports.values():
            # conn.shutdown is not coroutine
            conn.shutdown(reason)

    def _make_connection(self, sock, initiative, on_open=None, on_close=None):
        return self.transport_class(
            sock,
            initiative=initiative,
            ssl_context=self.ssl_context,
            ssl_handshake_timeout=self.ssl_handshake_timeout,
            on_open=on_open,
            on_close=on_close,
        )


ChannelType = TypeVar('ChannelType', bound=Channel)



import ssl as _ssl

from typing import Optional, Tuple, Type, TypeVar, Union

from pymaid.ext.middleware import MiddlewareManager
from pymaid.net.channel import StreamChannel as NetStreamChannel
from pymaid.net.protocol import Protocol, ProtocolType
from pymaid.utils.logger import get_logger

from .connection import Connection, ConnectionType
from .handler import Handler, SerialHandler
from .router import Router
from .types import HandlerType, RouterType

__all__ = ('Channel',)

logger = get_logger('rpc')


class Channel(NetStreamChannel):
    '''Channel manage connections.

    RPC *mostly* rely on ordered data, so it should be in stream mode.
    '''

    logger = logger

    def __init__(
        self,
        *,
        name: str = 'RPCChannel',
        address: Union[Tuple[str, int], str] = '',
        transport_class: ConnectionType = Connection,
        ssl_context: _ssl.SSLContext,
        ssl_handshake_timeout: Optional[float] = None,
        close_conn_onerror: bool = True,
        protocol: ProtocolType = Protocol,
        handler_class: Type[Handler] = SerialHandler,
        router_class: Type[Router] = Router,
        middleware_manager: Optional[MiddlewareManager] = None,
    ):
        super().__init__(
            name=name,
            address=address,
            transport_class=transport_class,
            ssl_context=ssl_context,
            ssl_handshake_timeout=ssl_handshake_timeout,
            middleware_manager=middleware_manager,
        )

        self.protocol = protocol
        self.handler_class = handler_class
        self.router = router_class()

    def _make_connection(
        self,
        sock,
        initiative,
        on_open=None,
        on_close=None,
    ) -> ConnectionType:
        return self.transport_class(
            sock,
            initiative=initiative,
            ssl_context=self.ssl_context,
            ssl_handshake_timeout=self.ssl_handshake_timeout,
            on_open=on_open,
            on_close=on_close,
            protocol=self.protocol,
            handler=self.handler_class(self.router),
        )

    def connection_made(self, sock) -> ConnectionType:
        conn = super().connection_made(sock)
        self.middleware_manager.dispatch('on_connection_made', self, conn)
        return conn

    def connection_lost(
        self,
        conn: ConnectionType,
        exc: Optional[Exception] = None
    ):
        # full = self.is_full
        super().connection_lost(conn, exc)
        self.middleware_manager.dispatch('on_connection_lost', self, conn)
        # if (full
        #         and not self.is_full
        #         and not self.is_paused
        #         and not self.is_stopped
        #         and self.middleware_manager.dispatch('can_accept', self)):
        #     self.start()

    def start(self):
        super().start()
        self.middleware_manager.dispatch('on_start', self)

    def pause(self, reason: str = ''):
        super().pause(reason)
        self.middleware_manager.dispatch('on_pause', self)

    def shutdown(self, reason: str = 'shutdown'):
        super().shutdown(reason)
        self.middleware_manager.dispatch('on_shutdown', self)

    def close(
        self, reason: Union[None, str, Exception] = 'called close',
    ):
        super().close(reason)
        self.middleware_manager.dispatch('on_close', self)


ChannelType = TypeVar('Channel', bound=Channel)


def create_stream_channel(
    address: Union[Tuple[str, int], str],
    *,
    name: str = 'StreamChannel',
    transport_class: ConnectionType = Connection,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    middleware_manager: Optional[MiddlewareManager] = None,
    close_conn_onerror: bool = True,
    protocol: Protocol = Protocol,
    handler_class: Optional[HandlerType] = SerialHandler,
    router_class: Optional[RouterType] = Router,
) -> ChannelType:
    return Channel(
        address=address,
        name=name,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        middleware_manager=middleware_manager,
        close_conn_onerror=close_conn_onerror,
        protocol=protocol,
        handler_class=handler_class,
        router_class=router_class,
    )



from . import conf
from . import error
from . import net
from . import utils

from .core import *  # noqa

__all__ = ('conf', 'error', 'net', 'utils')

__version__ = '1.0.0a1'

if not conf.settings.get('NO_UVLOOP', ns='pymaid'):
    import uvloop
    uvloop.install()



import pymaid
from pymaid.net.stream import Stream

from examples.template import get_server_parser, parse_args


class Stream(Stream):

    def data_received(self, data):
        # cannot use asynchronous way since this is in io callback
        self.write_sync(data)


async def main(args):
    ch = await pymaid.net.serve_stream(args.address, transport_class=Stream)
    async with ch:
        await ch.serve_forever()


if __name__ == "__main__":
    args = parse_args(get_server_parser())
    pymaid.run(main(args))



from pymaid.error.base import ErrorManager


RPCError = ErrorManager.create_manager('RPCError')
RPCError.add_error('ServerPaused', 'server pause accept new connection')
RPCError.add_error('ConnectionLimit', 'server is full')
RPCError.add_error('RPCRequestSent', '')
RPCError.add_error('RPCRequestReceived', '')
RPCError.add_error('RPCResponseSent', '')
RPCError.add_error('RPCResponseReceived', '')
RPCError.add_error('MultipleRequestForUnaryMethod', '')
RPCError.add_error('MultipleResponseForUnaryMethod', '')
RPCError.add_error('RPCShutdown', '')
